# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

"""
**Module containing all the elements to track the RF frequency, voltage, phase,
and the beam coordinates in phase space.**

:Authors:  **Helga Timko**, **Alexandre Lasheen**, **Danilo Quartullo**
"""

import warnings
from builtins import range

import numpy as np
import scipy
from packaging.version import Version

if Version(scipy.__version__) >= Version("1.14"):
    from scipy.integrate import cumulative_trapezoid as cumtrapz
else:
    from scipy.integrate import cumtrapz

from ..utils import bmath as bm


class FullRingAndRF:
    """
    *Definition of the full ring and RF parameters in order to be able to have
    a full turn information (used in the hamiltonian for example).*
    """

    def __init__(self, RingAndRFSection_list):

        #: *List of the total RingAndRFSection objects*
        self.RingAndRFSection_list = RingAndRFSection_list

        #: *Total potential well in [V]*
        self.potential_well = 0

        #: *Total potential well theta coordinates in [rad]*
        self.potential_well_coordinates = 0

        #: *Ring circumference in [m]*
        self.ring_circumference = 0
        for RingAndRFSectionElement in self.RingAndRFSection_list:
            self.ring_circumference += RingAndRFSectionElement.rf_params.section_length

        #: *Ring radius in [m]*
        self.ring_radius = self.ring_circumference / (2 * np.pi)

    def potential_well_generation(self, turn=0, n_points=int(1e5),
                                  main_harmonic_option='lowest_freq',
                                  dt_margin_percent=0., time_array=None):
        """Method to generate the potential well out of the RF systems. The
        assumption made is that all the RF voltages are averaged over one turn.
        The potential well is then approximated over one turn, which is not the
        exact potential. This approximation should be fine enough to generate a
        bunch (the mismatch should be small and damped fast enough). The
        default main harmonic is defined to be the lowest one in frequency. The
        user can change this option if it is not the case for his simulations
        (other options are: 'highest_voltage', or inputing directly the value
        of the desired main harmonic). A margin on the time array can be
        applied in order to be able to see the min/max that might be exactly on
        the edges of the frame (by adding a % to the length of the frame, this
        is set to 0 by default. It assumes also that the slippage factor is the
        same in the whole ring.
        """

        voltages = np.array([])
        omega_rf = np.array([])
        phi_offsets = np.array([])

        for RingAndRFSectionElement in self.RingAndRFSection_list:
            RF_params = RingAndRFSectionElement.rf_params
            charge = RF_params.Particle.charge
            for rf_system in range(RF_params.n_rf):
                voltages = np.append(voltages,
                                     RF_params.voltage[rf_system, turn])
                omega_rf = np.append(omega_rf,
                                     RF_params.omega_rf[rf_system, turn])
                phi_offsets = np.append(phi_offsets,
                                        RF_params.phi_rf[rf_system, turn])

        voltages = np.array(voltages, ndmin=2)
        omega_rf = np.array(omega_rf, ndmin=2)
        phi_offsets = np.array(phi_offsets, ndmin=2)

        if main_harmonic_option == 'lowest_freq':
            main_omega_rf = np.min(omega_rf)
        elif main_harmonic_option == 'highest_voltage':
            main_omega_rf = np.min(omega_rf[voltages == np.max(voltages)])
        elif isinstance(main_harmonic_option, int) or \
                isinstance(main_harmonic_option, float):
            if omega_rf[omega_rf == main_harmonic_option].size == 0:
                # PotentialWellError
                raise RuntimeError("ERROR in FullRingAndRF: The desired" +
                                   " harmonic to compute the potential well does not match" +
                                   " the RF parameters...")
            main_omega_rf = np.min(omega_rf[omega_rf == main_harmonic_option])

        slippage_factor = self.RingAndRFSection_list[0].rf_params.eta_0[turn]

        if time_array is None:
            time_array_margin = dt_margin_percent * 2 * np.pi / main_omega_rf

            first_dt = - time_array_margin / 2
            last_dt = 2 * np.pi / main_omega_rf + time_array_margin / 2

            time_array = np.linspace(float(first_dt), float(last_dt), int(n_points))

        self.total_voltage = np.sum(voltages.T *
                                    np.sin(omega_rf.T * time_array + phi_offsets.T), axis=0)

        eom_factor_potential = np.sign(slippage_factor) * charge / \
            (RingAndRFSectionElement.rf_params.t_rev[turn])

        potential_well = - cumtrapz(eom_factor_potential * (self.total_voltage -
                                                            (- RingAndRFSectionElement.acceleration_kick[turn]) / abs(charge)),
                                    dx=time_array[1] - time_array[0], initial=0)
        potential_well = potential_well - np.min(potential_well)

        self.potential_well_coordinates = time_array
        self.potential_well = potential_well

    def track(self):
        """Function to loop over all the RingAndRFSection.track methods
        """

        for RingAndRFSectionElement in self.RingAndRFSection_list:
            RingAndRFSectionElement.track()


class RingAndRFTracker:
    r""" Class taking care of basic particle coordinate tracking for a given
    RF station and the part of the ring until the next station, see figure.

    .. image:: ring_and_RFstation.png
        :align: center
        :width: 600
        :height: 600

    The time step is fixed to be one turn, but the tracking can consist of
    multiple RingAndRFTracker objects. In this case, the user should make sure
    that the lengths of the stations sum up exactly to the circumference or use
    the FullRingAndRF object in order to let the code pre-process the
    parameters. Each RF station may contain several RF harmonic systems which
    are considered to be in the same location. First, the energy kick of the RF
    station is applied, and then the particle arrival time to the next station
    is updated. The change in RF phase, voltage, and frequency due to control
    loops is tracked as well.

    Parameters
    ----------
    RFStation : class
        A RFStation type class
    counter : [int]
        Inherited from
        :py:attr:`input_parameters.rf_parameters.RFStation.counter`
    acceleration_kick : float array
        Inherited from
        :py:attr:`input_parameters.ring.Ring.delta_E`
        and multiplied by -1
    Beam : class
        A Beam type class
    solver : str
        Type of solver used for the drift equation; use 'simple' for 1st order
        approximation and 'exact' for exact solver
    BeamFeedback : class (optional)
        A BeamFeedback type class, beam-based feedback on RF frequency;
        default is None
    NoiseFeedback : class (optional)
        A NoiseFeedback type class, bunch-length feedback on RF noise;
        default is None
    CavityFeedback : class or list of classes (optional)
        A CavityFeedback type child class, cavity feedback modulating the
        RF voltage bucket-by-bucket in amplitude and phase. Can either be a single object or a list of such objects.
        Default is None.
    periodicity : bool (optional)
        Option to switch periodic solver on/off; default is False (off)
    interpolation : bool (optional)
        Option to use sliced and interpolated voltage for the kicker; default
        is False

    """

    def __init__(self, RFStation, Beam, solver='simple', BeamFeedback=None,
                 NoiseFeedback=None, CavityFeedback=None, periodicity=False,
                 interpolation=False, Profile=None, TotalInducedVoltage=None):

        # Set up logging
        # self.logger = logging.getLogger(__class__.__name__)
        # self.logger.info("Class initialized")

        # Imports from RF parameters
        self.rf_params = RFStation
        self.counter = RFStation.counter
        self.acceleration_kick = - RFStation.delta_E

        # Other imports
        self.beam = Beam
        self.solver = str(solver)
        if self.solver not in ['simple', 'exact', 'legacy']:
            # SolverError
            raise RuntimeError("ERROR in RingAndRFTracker: Choice of" +
                               " longitudinal solver not recognised!")
        if self.rf_params.alpha_order > 1:  # Force exact solver for higher orders of eta
            self.solver = 'exact'

        # Options
        self.beamFB = BeamFeedback
        self.noiseFB = NoiseFeedback
        self.cavityFB = CavityFeedback
        try:
            self.periodicity = bool(periodicity)
        except Exception:
            # PeriodicityError
            raise RuntimeError("ERROR in RingAndRFTracker: Choice of" +
                               " periodicity not recognised!")
        try:
            self.interpolation = bool(interpolation)
        except Exception:
            # InterpolationError
            raise RuntimeError("ERROR in RingAndRFTracker: Choice of" +
                               " interpolation not recognised!")
        self.profile = Profile
        self.totalInducedVoltage = TotalInducedVoltage
        if (self.interpolation is True) and (self.profile is None):
            # ProfileError
            raise RuntimeError("ERROR in RingAndRFTracker: Please specify a" +
                               " Profile object to use the interpolation option")

        if (self.cavityFB is not None) and (self.profile is None):
            # ProfileError
            raise RuntimeError("ERROR in RingAndRFTracker: Please specify a" +
                               " Profile object to use the CavityFeedback class")
        if (self.rf_params.empty is True) and (self.periodicity is True):
            # PeriodicityError
            raise RuntimeError("ERROR in RingAndRFTracker: Empty RFStation" +
                               " with periodicity not yet implemented!")
        if (self.cavityFB is not None) and (self.interpolation is False):
            self.interpolation = True
            warnings.warn('Setting interpolation to TRUE')
            # self.logger.warning("Setting interpolation to TRUE")

        if (self.cavityFB is not None) and (not hasattr(self.cavityFB, '__iter__')):
            self.cavityFB = [self.cavityFB]

    def kick(self, beam_dt, beam_dE, index):
        r"""Function updating the particle energy due to the RF kick in a given
        RF station. The kicks are summed over the different harmonic RF systems
        in the station. The cavity phase can be shifted by the user via
        phi_offset. The main RF (harmonic[0]) has by definition phase = 0 at
        time = 0 below transition. The phases of all other RF systems are
        defined w.r.t.\ to the main RF. The increment in energy is given by the
        discrete equation of motion:

        .. math::
            \Delta E^{n+1} = \Delta E^n + \sum_{k=0}^{n_{\mathsf{rf}}-1}{e V_k^n \\sin{\\left(\omega_{\mathsf{rf,k}}^n \\Delta t^n + \phi_{\mathsf{rf,k}}^n \\right)}} - (E_s^{n+1} - E_s^n)

        """

        bm.kick(beam_dt, beam_dE, self.rf_params.voltage[:, index],
                self.rf_params.omega_rf[:, index], self.rf_params.phi_rf[:, index],
                self.rf_params.Particle.charge, self.rf_params.n_rf, self.acceleration_kick[index])

    def drift(self, beam_dt, beam_dE, index):
        r"""Function updating the particle arrival time to the RF station
        (drift). If only the zeroth order slippage factor is given, 'simple'
        and 'exact' solvers are available. The 'simple' solver is somewhat
        faster. Otherwise, the solver is automatically 'exact' and calculates
        the frequency slippage up to second order. The corresponding equations
        are (nb: the n indices correspond to the turn number):

        .. math::
            \\Delta t^{n+1} = \\Delta t^{n} + \\frac{L}{C} T_0^{n+1} \\left[ \\left(1+\\sum_{i=0}^{2}{\\alpha_i\\left(\\delta^{n+1}\\right)^{i+1}}\\right)   \\frac{1+\\left(\\Delta E/E_s\\right)^{n+1}}{1+\\delta^{n+1}}    - 1\\right] \quad \\text{(exact)}

        .. math::
            \\Delta t^{n+1} = \\Delta t^{n} + \\frac{L}{C} T_0^{n+1} \\left(\\frac{1}{1 - \\eta(\\delta^{n+1})\\delta^{n+1}} - 1\\right) \quad \\text{(legacy)}

        .. math::
            \\Delta t^{n+1} = \\Delta t^{n} + \\frac{L}{C} T_0^{n+1}\\eta_0\\delta^{n+1} \quad \\text{(simple)}

        The relative momentum needs to be calculated from the relative energy
        and is obtained as follow:

        .. math::
            \\delta = \\sqrt{1+\\beta_s^{-2}\\left[\\left(\\frac{\\Delta E}{E_s}\\right)^2 + 2\\frac{\\Delta E}{E_s}\\right]} - 1 \quad \\text{(exact)}

        .. math::
            \\delta = \\frac{\\Delta E}{\\beta_s^2 E_s} \quad \\text{(simple, legacy)}

        """
        bm.drift(beam_dt, beam_dE, self.solver, self.rf_params.t_rev[index],
                 self.rf_params.length_ratio, self.rf_params.alpha_order, self.rf_params.eta_0[index],
                 self.rf_params.eta_1[index], self.rf_params.eta_2[index], self.rf_params.alpha_0[index],
                 self.rf_params.alpha_1[index], self.rf_params.alpha_2[index],
                 self.rf_params.beta[index], self.rf_params.energy[index])

    def rf_voltage_calculation(self):
        """Function calculating the total, discretised RF voltage seen by the
        beam at a given turn. Requires a Profile object.

        """
        voltages = bm.ascontiguousarray(self.rf_params.voltage[:, self.counter[0]])
        omega_rf = bm.ascontiguousarray(self.rf_params.omega_rf[:, self.counter[0]])
        phi_rf = bm.ascontiguousarray(self.rf_params.phi_rf[:, self.counter[0]])
        # TODO: test with multiple harmonics, think about 800 MHz OTFB
        if self.cavityFB:
            # Allocate memory for rf_voltage and reset it to zero
            self.rf_voltage = np.zeros(self.profile.n_slices)

            # Add corrections from cavity feedbacks for the different harmonics
            for ind, feedback in enumerate(self.cavityFB):
                if feedback is not None:
                    self.rf_voltage += voltages[ind] * feedback.V_corr * \
                                      bm.sin(omega_rf[ind] * self.profile.bin_centers +
                                             phi_rf[ind] + feedback.phi_corr)
                else:
                    self.rf_voltage += bm.rf_volt_comp(voltages[ind:ind + 1], omega_rf[ind:ind + 1],
                                                       phi_rf[ind:ind + 1], self.profile.bin_centers)

            # Add RF voltage from harmonics that do not have a cavity feedback model
            self.rf_voltage += bm.rf_volt_comp(voltages[len(self.cavityFB):],
                                               omega_rf[len(self.cavityFB):],
                                               phi_rf[len(self.cavityFB):],
                                               self.profile.bin_centers)
        else:
            self.rf_voltage = bm.rf_volt_comp(voltages, omega_rf, phi_rf,
                                              self.profile.bin_centers)

    def track(self):
        """Tracking method for the section. Applies first the kick, then the
        drift. Calls also RF/beam feedbacks if applicable. Updates the counter
        of the corresponding RFStation class and the energy-related variables
        of the Beam class.

        """
        turn = self.counter[0]

        # Add phase noise directly to the cavity RF phase
        if self.rf_params.phi_noise is not None:
            if self.noiseFB is not None:
                self.rf_params.phi_rf[:, turn] += \
                    self.noiseFB.x * self.rf_params.phi_noise[:, turn]
            else:
                self.rf_params.phi_rf[:, turn] += \
                    self.rf_params.phi_noise[:, turn]

        # Add phase modulation directly to the cavity RF phase
        if self.rf_params.phi_modulation is not None:
            self.rf_params.phi_rf[:, turn] += \
                self.rf_params.phi_modulation[0][:, turn]
            self.rf_params.omega_rf[:, turn] += \
                self.rf_params.phi_modulation[1][:, turn]

        # Determine phase loop correction on RF phase and frequency
        if self.beamFB is not None and turn >= self.beamFB.delay:
            self.beamFB.track()

        # Update the RF phase of all systems for the next turn
        # Accumulated phase offset due to beam phase loop or frequency offset
        self.rf_params.dphi_rf += 2. * np.pi * self.rf_params.harmonic[:, turn + 1] * \
            (self.rf_params.omega_rf[:, turn + 1] -
             self.rf_params.omega_rf_d[:, turn + 1]) / \
            self.rf_params.omega_rf_d[:, turn + 1]

        # Total phase offset
        self.rf_params.phi_rf[:, turn + 1] += self.rf_params.dphi_rf

        # Correction from cavity loop
        if self.cavityFB is not None:
            for feedback in self.cavityFB:
                if feedback is not None:
                    feedback.track()

        if self.periodicity:

            # Distinguish the particles inside the frame from the particles on
            # the right-hand side of the frame.
            self.indices_right_outside = \
                bm.where(self.beam.dt > self.rf_params.t_rev[turn + 1])[0]
            self.indices_inside_frame = \
                bm.where(self.beam.dt < self.rf_params.t_rev[turn + 1])[0]

            if len(self.indices_right_outside) > 0:
                # Change reference of all the particles on the right of the
                # current frame; these particles skip one kick and drift
                self.beam.dt[self.indices_right_outside] -= \
                    self.rf_params.t_rev[turn + 1]
                # Synchronize the bunch with the particles that are on the
                # RHS of the current frame applying kick and drift to the
                # bunch
                # After that all the particles are in the new updated frame
                self.insiders_dt = bm.ascontiguousarray(
                    self.beam.dt[self.indices_inside_frame])
                self.insiders_dE = bm.ascontiguousarray(
                    self.beam.dE[self.indices_inside_frame])
                self.kick(self.insiders_dt, self.insiders_dE, turn)
                self.drift(self.insiders_dt, self.insiders_dE, turn + 1)
                self.beam.dt[self.indices_inside_frame] = self.insiders_dt
                self.beam.dE[self.indices_inside_frame] = self.insiders_dE
                # Check all the particles on the left of the just updated
                # frame and apply a second kick and drift to them with the
                # previous wave after having changed reference.
                self.indices_left_outside = bm.where(self.beam.dt < 0)[0]

            else:
                self.kick(self.beam.dt, self.beam.dE, turn)
                self.drift(self.beam.dt, self.beam.dE, turn + 1)
                # Check all the particles on the left of the just updated
                # frame and apply a second kick and drift to them with the
                # previous wave after having changed reference.
                self.indices_left_outside = bm.where(self.beam.dt < 0)[0]

            if len(self.indices_left_outside) > 0:
                left_outsiders_dt = bm.ascontiguousarray(
                    self.beam.dt[self.indices_left_outside])
                left_outsiders_dE = bm.ascontiguousarray(
                    self.beam.dE[self.indices_left_outside])
                left_outsiders_dt += self.rf_params.t_rev[turn + 1]
                self.kick(left_outsiders_dt, left_outsiders_dE, turn)
                self.drift(left_outsiders_dt, left_outsiders_dE, turn + 1)
                self.beam.dt[self.indices_left_outside] = left_outsiders_dt
                self.beam.dE[self.indices_left_outside] = left_outsiders_dE

        else:

            if self.rf_params.empty is False:
                if self.interpolation:
                    self.rf_voltage_calculation()
                    if self.totalInducedVoltage is not None:
                        self.total_voltage = self.rf_voltage \
                            + self.totalInducedVoltage.induced_voltage
                    else:
                        self.total_voltage = self.rf_voltage

                    bm.linear_interp_kick(dt=self.beam.dt, dE=self.beam.dE,
                                          voltage=self.total_voltage,
                                          bin_centers=self.profile.bin_centers,
                                          charge=self.beam.Particle.charge,
                                          acceleration_kick=self.acceleration_kick[turn])

                else:
                    self.kick(self.beam.dt, self.beam.dE, turn)

            self.drift(self.beam.dt, self.beam.dE, turn + 1)

        # Updating the beam synchronous momentum etc.
        self.beam.beta = self.rf_params.beta[turn + 1]
        self.beam.gamma = self.rf_params.gamma[turn + 1]
        self.beam.energy = self.rf_params.energy[turn + 1]
        self.beam.momentum = self.rf_params.momentum[turn + 1]

        # Increment by one the turn counter
        self.counter[0] += 1

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return

        # transfer recursively objects
        if recursive and self.profile:
            self.profile.to_gpu()
        if recursive and self.totalInducedVoltage:
            self.totalInducedVoltage.to_gpu()
        if recursive and self.beam:
            self.beam.to_gpu()
        if recursive and self.beamFB:
            self.beamFB.to_gpu()
        if recursive and self.rf_params:
            self.rf_params.to_gpu()


        if hasattr(self, 'rf_voltage'):
            self.rf_voltage = self.rf_params.rf_voltage

        # to make sure it will not be called again
        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return

        # transfer recursively objects
        if recursive and self.profile:
            self.profile.to_cpu()
        if recursive and self.totalInducedVoltage:
            self.totalInducedVoltage.to_cpu()
        if recursive and self.beam:
            self.beam.to_cpu()
        if recursive and self.beamFB:
            self.beamFB.to_cpu()
        if recursive and self.rf_params:
            self.rf_params.to_cpu()

        import cupy as cp

        if hasattr(self, 'rf_voltage'):
            self.rf_voltage = cp.asnumpy(self.rf_voltage)

        # to make sure it will not be called again
        self._device = 'CPU'



# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module gathering all general input parameters used for the simulation.**
    :Authors: **Alexandre Lasheen**, **Danilo Quartullo**, **Helga Timko**
'''

from __future__ import division

import warnings
from builtins import range, str

import numpy as np
from scipy.constants import c

from ..input_parameters.ring_options import RingOptions


class Ring:
    r""" Class containing the general properties of the synchrotron that are
    independent of the RF system or the beam.

    The index :math:`n` denotes time steps, :math:`k` ring segments/sections
    and :math:`i` momentum compaction orders.

    Parameters
    ----------
    ring_length : float (opt: float array [n_sections])
        Length [m] of the n_sections ring segments of the synchrotron.
        An RF station, a synchrotron radiation kick, and/or an impedance kick
        can be included at the end of each ring section.
    alpha_0 : float (opt: float array/matrix [n_sections, n_turns+1])
        Momentum compaction factor of zeroth order :math:`\alpha_{0,k,i}` [1];
        can be input as single float or as a program of (n_turns + 1) turns
        (should be of the same size as synchronous_data).
        In case of higher order momentum compaction, check the
        documentation for the inputs: alpha_order, alpha_1, alpha_2
    synchronous_data : float (opt: float array/matrix [n_sections, n_turns+1])
        Design synchronous particle momentum (default) [eV], kinetic or
        total energy [eV] or bending field [T] on the design orbit.
        Input for each RF section :math:`p_{s,k,n}`.
        Can be input as a single constant float, or as a
        program of (n_turns + 1) turns. In case of several sections without
        acceleration, input: [[momentum_section_1], [momentum_section_2],
        etc.]. In case of several sections with acceleration, input:
        [momentum_program_section_1, momentum_program_section_2, etc.]. Can
        be input also as a tuple of time and momentum, see also
        'cycle_time' and 'PreprocessRamp'
    Particle : class
        A Particle-based class defining the primary, synchronous particle (mass
        and charge) that is reference for the momentum/energy in the ring.
    n_turns : int
        Optional: Number of turns :math:`n` [1] to be simulated.
        If a synchronous_data program is passed as a tuple (see below),
        the number of turns will be overwritten depending on the length in time
        of the program
    synchronous_data_type : str
        Optional: Choice of 'synchronous_data' type; can be 'momentum'
        (default), 'total energy', 'kinetic energy' or 'bending field'
        (requires bending_radius to be defined)
    bending_radius : float
        Optional: Radius [m] of the bending magnets,
        required if 'bending field' is set for the synchronous_data_type
    n_sections : int
        Optional: number of ring sections/segments; default is 1
    alpha_1 : float (opt: float array/matrix [n_sections, n_turns+1])
        Momentum compaction factor of first order
        :math:`\alpha_{1,k,i}` [1]; can be input as single float or as a
        program of (n_turns + 1) turns (should be of the same size as
        synchronous_data and alpha_0).
    alpha_2 : float (opt: float array/matrix [n_sections, n_turns+1])
        Optional : Momentum compaction factor of second order
        :math:`\alpha_{2,k,i}` [1]; can be input as single float or as a
        program of (n_turns + 1) turns (should be of the same size as
        synchronous_data and alpha_0).
    RingOptions : class
        Optional : A RingOptions-based class with default options to check the
        input and initialize the momentum program for the simulation.
        This object defines the interpolation scheme, plotting options, etc.
        The options for this object can be adjusted and passed to the Ring
        object.

    Attributes
    ----------
    ring_circumference : float
        Circumference of the synchrotron. Sum of ring segment lengths,
        :math:`C = \sum_k L_k` [m]
    ring_radius : float
        Radius of the synchrotron, :math:`R = C/(2 \pi)` [m]
    bending_radius : float
        Bending radius in dipole magnets, :math:`\rho` [m]
    alpha_order : int
        Highest order of momentum compaction (as defined by the input). Can
        be 0,1,2.
    alpha_0 : float matrix [n_sections, n_turns+1]
        Zeroth order momentum compaction factor
        :math:`\alpha_{0,k,n}`
    alpha_1 : float matrix [n_sections, n_turns+1]
        First order momentum compaction factor
        :math:`\alpha_{1,k,n}`
    alpha_2 : float matrix [n_sections, n_turns+1]
        Second order momentum compaction factor
        :math:`\alpha_{2,k,n}`
    eta_0 : float matrix [n_sections, n_turns+1]
        Zeroth order slippage factor :math:`\eta_{0,k,n} = \alpha_{0,k,n} -
        \frac{1}{\gamma_{s,k,n}^2}` [1]
    eta_1 : float matrix [n_sections, n_turns+1]
        First order slippage factor :math:`\eta_{1,k,n} =
        \frac{3\beta_{s,k,n}^2}{2\gamma_{s,k,n}^2} + \alpha_{1,k,n} -
        \alpha_{0,k,n}\eta_{0,k,n}` [1]
    eta_2 : float matrix [n_sections, n_turns+1]
        Second order slippage factor :math:`\eta_{2,k,n} =
        -\frac{\beta_{s,k,n}^2\left(5\beta_{s,k,n}^2-1\right)}
        {2\gamma_{s,k,n}^2} + \alpha_{2,k,n} - 2\alpha_{0,k,n}\alpha_{1,k,n}
        + \frac{\alpha_{1,k,n}}{\gamma_{s,k,n}^2} + \alpha_{0,k}^2\eta_{0,k,n}
        - \frac{3\beta_{s,k,n}^2\alpha_{0,k,n}}{2\gamma_{s,k,n}^2}` [1]
    momentum : float matrix [n_sections, n_turns+1]
        Synchronous relativistic momentum on the design orbit :math:`p_{s,k,n}`
    beta : float matrix [n_sections, n_turns+1]
        Synchronous relativistic beta program for each segment of the
        ring :math:`\beta_{s,k}^n = \frac{1}{\sqrt{1
        + \left(\frac{m}{p_{s,k,n}}\right)^2} }` [1]
    gamma : float matrix [n_sections, n_turns+1]
        Synchronous relativistic gamma program for each segment of the ring
        :math:`\gamma_{s,k,n} = \sqrt{ 1
        + \left(\frac{p_{s,k,n}}{m}\right)^2 }` [1]
    energy : float matrix [n_sections, n_turns+1]
        Synchronous total energy program for each segment of the ring
        :math:`E_{s,k,n} = \sqrt{ p_{s,k,n}^2 + m^2 }` [eV]
    kin_energy : float matrix [n_sections, n_turns+1]
        Synchronous kinetic energy program for each segment of the ring
        :math:`E_{s,kin} = \sqrt{ p_{s,k,n}^2 + m^2 } - m` [eV]
    delta_E : float matrix [n_sections, n_turns+1]
        Derivative of synchronous total energy w.r.t. time, for all sections,
        :math:`: \quad E_{s,k,n+1}- E_{s,k,n}` [eV]
    t_rev : float array [n_turns+1]
        Revolution period turn by turn.
        :math:`T_{0,n} = \frac{C}{\beta_{s,n} c}` [s]
    f_rev : float array [n_turns+1]
        Revolution frequency :math:`f_{0,n} = \frac{1}{T_{0,n}}` [Hz]
    omega_rev : float array [n_turns+1]
        Revolution angular frequency :math:`\omega_{0,n} = 2\pi f_{0,n}` [1/s]
    cycle_time : float array [n_turns+1]
        Cumulative cycle time, turn by turn, :math:`t_n = \sum_n T_{0,n}` [s].
        Possibility to extract cycle parameters at these moments using
        'parameters_at_time'.
    RingOptions : RingOptions()
        The RingOptions is kept as an attribute of the Ring object for further
        usage.

    Examples
    --------
    >>> # To declare a single-section synchrotron at constant energy:
    >>> # Particle type Proton
    >>> from blond.beam.beam import Proton
    >>> from blond.input_parameters.ring import Ring
    >>>
    >>> n_turns = 10
    >>> C = 26659
    >>> alpha_0 = 3.21e-4
    >>> momentum = 450e9
    >>> ring = Ring(C, alpha_0, momentum, Proton(), n_turns)
    >>>
    >>>
    >>> # To declare a single-section synchrotron with a momentum program in time:
    >>> # Particle type Proton
    >>> from blond.beam.beam import Proton
    >>> from blond.input_parameters.ring import Ring
    >>>
    >>> C = 26659
    >>> alpha_0 = 3.21e-4
    >>> momentum_time = (0, 500e-3)
    >>> momentum = (450e9, 451e9)
    >>> momentum_program = (momentum_time, momentum)
    >>> ring = Ring(C, alpha_0, momentum_program, Proton())
    >>> # NB: n_turns is overwritten with the actual turns for the defined program
    >>>
    >>>
    >>> # To declare a two section synchrotron at constant energy and
    >>> # higher-order momentum compaction factors; particle Electron:
    >>> from blond.beam.beam import Electron
    >>> from blond.input_parameters.ring import Ring
    >>>
    >>> n_turns = 10
    >>> C = [13000, 13659]
    >>> alpha_0 = [[3.21e-4], [2.89e-4]]  # or [3.21e-4, 2.89e-4]
    >>> alpha_1 = [[2.e-5], [1.e-5]]  # or [2.e-5, 1.e-5]
    >>> alpha_2 = [[5.e-7], [5.e-7]]  # or [5.e-7, 5.e-7]
    >>> momentum = 450e9
    >>> ring = Ring(C, alpha_0, momentum, Electron(), n_turns,
    >>>             alpha_1=alpha_1, alpha_2=alpha_2)

    """

    def __init__(self, ring_length, alpha_0, synchronous_data, Particle,
                 n_turns=1, synchronous_data_type='momentum',
                 bending_radius=None, n_sections=1, alpha_1=None, alpha_2=None,
                 RingOptions=RingOptions()):

        # Conversion of initial inputs to expected types
        self.n_turns = int(n_turns)
        self.n_sections = int(n_sections)

        # Ring length and checks
        self.ring_length = np.array(ring_length, ndmin=1, dtype=float)
        self.ring_circumference = np.sum(self.ring_length)
        self.ring_radius = self.ring_circumference / (2 * np.pi)

        if bending_radius is not None:
            self.bending_radius = float(bending_radius)
        else:
            self.bending_radius = bending_radius

        if self.n_sections != len(self.ring_length):
            # InputDataError
            raise RuntimeError("ERROR in Ring: Number of sections and ring " +
                               "length size do not match!")

        # Primary particle mass and charge used for energy calculations
        self.Particle = Particle

        # Keeps RingOptions as an attribute
        self.RingOptions = RingOptions

        # Reshaping the input synchronous data to the adequate format and
        # get back the momentum program from RingOptions
        self.momentum = RingOptions.reshape_data(
            synchronous_data,
            self.n_turns,
            self.n_sections,
            input_to_momentum=True,
            synchronous_data_type=synchronous_data_type,
            mass=self.Particle.mass,
            charge=self.Particle.charge,
            circumference=self.ring_circumference,
            bending_radius=self.bending_radius)

        # Updating the number of turns in case it was changed after ramp
        # interpolation
        if self.momentum.shape[1] != (self.n_turns + 1):
            self.n_turns = self.momentum.shape[1] - 1
            warnings.warn("WARNING in Ring: The number of turns for the " +
                          "simulation was changed by passing a momentum " +
                          "program.")

        # Derived from momentum
        self.beta = np.sqrt(1 / (1 + (self.Particle.mass / self.momentum)**2))
        self.gamma = np.sqrt(1 + (self.momentum / self.Particle.mass)**2)
        self.energy = np.sqrt(self.momentum**2 + self.Particle.mass**2)
        self.kin_energy = np.sqrt(self.momentum**2 + self.Particle.mass**2) - \
            self.Particle.mass
        self.delta_E = np.diff(self.energy, axis=1)
        self.t_rev = np.dot(self.ring_length, 1 / (self.beta * c))
        self.cycle_time = np.cumsum(self.t_rev)  # Always starts with zero
        self.f_rev = 1 / self.t_rev
        self.omega_rev = 2 * np.pi * self.f_rev

        # Momentum compaction, checks, and derived slippage factors
        if RingOptions.t_start is None:
            interp_time = self.cycle_time
        else:
            interp_time = self.cycle_time + RingOptions.t_start

        self.alpha_0 = RingOptions.reshape_data(
            alpha_0, self.n_turns, self.n_sections,
            interp_time=interp_time)
        self.alpha_order = 0

        if alpha_1 is not None:
            self.alpha_1 = RingOptions.reshape_data(
                alpha_1, self.n_turns, self.n_sections,
                interp_time=interp_time)
            self.alpha_order = 1
        else:
            # Filling alpha_1 with zeros
            # This can be removed when the BLonD assembler is in place
            # to avoid high order momentum compaction programs filled
            # with zeros (should be propagated in RFStation.__init__())
            self.alpha_1 = np.zeros(self.alpha_0.shape)

        if alpha_2 is not None:
            self.alpha_2 = RingOptions.reshape_data(
                alpha_2, self.n_turns, self.n_sections,
                interp_time=interp_time)
            self.alpha_order = 2
        else:
            # Filling alpha_2 with zeros
            # This can be removed when the BLonD assembler is in place
            # to avoid high order momentum compaction programs filled
            # with zeros (should be propagated in RFStation.__init__())
            self.alpha_2 = np.zeros(self.alpha_0.shape)

        # Slippage factor derived from alpha, beta, gamma
        self.eta_generation()

    def eta_generation(self):
        """ Function to generate the slippage factors (zeroth, first, and
        second orders, see [1]_) from the momentum compaction and the
        relativistic beta and gamma program through the cycle.

        References
        ----------
        .. [1] "Accelerator Physics," S. Y. Lee, World Scientific,
                Third Edition, 2012.
        """

        for i in range(self.alpha_order + 1):
            getattr(self, '_eta' + str(i))()

        # Fill unused eta arrays with zeros
        # This can be removed when the BLonD assembler is in place
        # to avoid high order momentum compaction programs filled
        # with zeros (should be propagated in RFStation.__init__())
        for i in range(self.alpha_order + 1, 3):
            setattr(self, "eta_%s" % i, np.zeros([self.n_sections,
                                                  self.n_turns + 1]))

    def _eta0(self):
        """ Function to calculate the zeroth order slippage factor eta_0 """

        self.eta_0 = np.empty([self.n_sections, self.n_turns + 1])
        for i in range(0, self.n_sections):
            self.eta_0[i] = self.alpha_0[i] - self.gamma[i]**(-2.)

    def _eta1(self):
        """ Function to calculate the first order slippage factor eta_1 """

        self.eta_1 = np.empty([self.n_sections, self.n_turns + 1])
        for i in range(0, self.n_sections):
            self.eta_1[i] = 3 * self.beta[i]**2 / (2 * self.gamma[i]**2) + \
                self.alpha_1[i] - self.alpha_0[i] * self.eta_0[i]

    def _eta2(self):
        """ Function to calculate the second order slippage factor eta_2 """

        self.eta_2 = np.empty([self.n_sections, self.n_turns + 1])
        for i in range(0, self.n_sections):
            self.eta_2[i] = - self.beta[i]**2 * (5 * self.beta[i]**2 - 1) / \
                (2 * self.gamma[i]**2) + self.alpha_2[i] - 2 * self.alpha_0[i] *\
                self.alpha_1[i] + self.alpha_1[i] / self.gamma[i]**2 + \
                self.alpha_0[i]**2 * self.eta_0[i] - 3 * self.beta[i]**2 * \
                self.alpha_0[i] / (2 * self.gamma[i]**2)

    def parameters_at_time(self, cycle_moments):
        """ Function to return various cycle parameters at a specific moment in
        time. The cycle time is defined to start at zero in turn zero.

        Parameters
        ----------
        cycle_moments : float array
            Moments of time at which cycle parameters are to be calculated [s].

        Returns
        -------
        parameters : dictionary
            Contains 'momentum', 'beta', 'gamma', 'energy', 'kin_energy',
            'f_rev', 't_rev'. 'omega_rev', 'eta_0', and 'delta_E' interpolated
            to the moments contained in the 'cycle_moments' array

        """

        parameters = {}
        parameters['momentum'] = np.interp(cycle_moments, self.cycle_time,
                                           self.momentum[0])
        parameters['beta'] = np.interp(cycle_moments, self.cycle_time,
                                       self.beta[0])
        parameters['gamma'] = np.interp(cycle_moments, self.cycle_time,
                                        self.gamma[0])
        parameters['energy'] = np.interp(cycle_moments, self.cycle_time,
                                         self.energy[0])
        parameters['kin_energy'] = np.interp(cycle_moments, self.cycle_time,
                                             self.kin_energy[0])
        parameters['f_rev'] = np.interp(cycle_moments, self.cycle_time,
                                        self.f_rev)
        parameters['t_rev'] = np.interp(cycle_moments, self.cycle_time,
                                        self.t_rev)
        parameters['omega_rev'] = np.interp(cycle_moments, self.cycle_time,
                                            self.omega_rev)
        parameters['eta_0'] = np.interp(cycle_moments, self.cycle_time,
                                        self.eta_0[0])
        parameters['delta_E'] = np.interp(cycle_moments,
                                          self.cycle_time[1:],
                                          np.diff(self.energy[0]))

        return parameters



# coding: utf-8
# Copyright 2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Fitting and filters routines to be used alone or with the Profile class in
the beam package.**

:Authors: **Danilo Quartullo**, **Alexandre Lasheen**,
          **Juan F. Esteban Mueller**
'''

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import curve_fit
from scipy.signal import cheb2ord, cheby2, filtfilt, freqz

from ..utils import bmath as bm


def beam_profile_filter_chebyshev(Y_array, X_array, filter_option):
    """
    This routine is filtering the beam profile with a type II Chebyshev
    filter. The input is a library having the following structure and
    informations:

    filter_option = {'type':'chebyshev', 'pass_frequency':pass_frequency,
    'stop_frequency':stop_frequency, 'gain_pass':gain_pass,
    'gain_stop':gain_stop}

    The function returns nCoefficients, the number of coefficients used
    in the filter. You can also add the following option to plot and return
    the filter transfer function:

    filter_option = {..., 'transfer_function_plot':True}
    """

    noisyProfile = np.array(Y_array)

    freqSampling = 1 / (X_array[1] - X_array[0])
    nyqFreq = freqSampling / 2.

    frequencyPass = filter_option['pass_frequency'] / nyqFreq
    frequencyStop = filter_option['stop_frequency'] / nyqFreq
    gainPass = filter_option['gain_pass']
    gainStop = filter_option['gain_stop']

    # Compute the lowest order for a Chebyshev Type II digital filter
    nCoefficients, wn = cheb2ord(frequencyPass, frequencyStop, gainPass,
                                 gainStop)

    # Compute the coefficients a Chebyshev Type II digital filter
    b, a = cheby2(nCoefficients, gainStop, wn, btype='low')

    # Apply the filter forward and backwards to cancel the group delay
    Y_array = filtfilt(b, a, noisyProfile)
    Y_array = np.ascontiguousarray(Y_array)

    if (('transfer_function_plot' in filter_option)
            and filter_option['transfer_function_plot']):
        # Plot the filter transfer function
        w, transferGain = freqz(b, a=a, worN=len(Y_array))
        transferFreq = w / np.pi * nyqFreq
        group_delay = -np.diff(-np.unwrap(-np.angle(transferGain))) / \
                      -np.diff(w * freqSampling)

        plt.figure()
        ax1 = plt.subplot(311)
        plt.plot(transferFreq, 20 * np.log10(abs(transferGain)))
        plt.ylabel('Magnitude [dB]')
        plt.subplot(312, sharex=ax1)
        plt.plot(transferFreq, np.unwrap(-np.angle(transferGain)))
        plt.ylabel('Phase [rad]')
        plt.subplot(313, sharex=ax1)
        plt.plot(transferFreq[:-1], group_delay)
        plt.ylabel('Group delay [s]')
        plt.xlabel('Frequency [Hz]')

        # Plot the bunch spectrum and the filter transfer function
        plt.figure()
        plt.plot(
            np.fft.fftfreq(len(Y_array), X_array[1] - X_array[0]),
            20. * np.log10(np.abs(np.fft.fft(noisyProfile))))
        plt.xlabel('Frequency [Hz]')
        plt.twinx()
        plt.plot(transferFreq, 20 * np.log10(abs(transferGain)), 'r')
        plt.xlim(0, plt.xlim()[1])

        plt.show()

        return Y_array

    else:

        return Y_array


def gaussian_fit(Y_array, X_array, p0):
    """
    Gaussian fit of the profile, in order to get the bunch length and
    position. Returns fit values in units of s.
    """
    if bm.device == 'GPU':
        X_array = X_array.get()
        Y_array = Y_array.get()

    return curve_fit(gauss, X_array, Y_array, p0)[0]


def gauss(x, *p):
    r"""
    Defined as:

    .. math:: A \, e^{\\frac{\\left(x-x_0\\right)^2}{2\\sigma_x^2}}

    """

    A, x0, sx = p
    return A * np.exp(-(x - x0)**2 / 2. / sx**2)


def rms(Y_array, X_array):
    """
    Computation of the RMS bunch length and position from the line
    density (bunch length = 4sigma).
    """

    timeResolution = X_array[1] - X_array[0]

    lineDenNormalized = Y_array / np.trapz(Y_array, dx=timeResolution)

    bp_rms = np.trapz(X_array * lineDenNormalized, dx=timeResolution)

    bl_rms = 4 * np.sqrt(
        np.trapz((X_array - bp_rms)**2 * lineDenNormalized, dx=timeResolution))

    return bp_rms, bl_rms


def fwhm(Y_array, X_array, shift=0):
    """
    Computation of the bunch length and position from the FWHM
    assuming Gaussian line density.
    """

    half_max = shift + 0.5 * (Y_array.max() - shift)

    # First aproximation for the half maximum values
    taux = np.where(Y_array >= half_max)
    t1 = taux[0][0]
    t2 = taux[0][-1]
    # Interpolation of the time where the line density is half the maximum
    bin_size = X_array[1] - X_array[0]
    try:
        t_left = X_array[t1] - bin_size * \
            (Y_array[t1] - half_max) / \
            (Y_array[t1] - Y_array[t1 - 1])
        t_right = X_array[t2] + bin_size * \
            (Y_array[t2] - half_max) / \
            (Y_array[t2] - Y_array[t2 + 1])

        bl_fwhm = 4 * (t_right - t_left) / (2 * np.sqrt(2 * np.log(2)))
        bp_fwhm = (t_left + t_right) / 2
    except Exception:
        bl_fwhm = np.nan
        bp_fwhm = np.nan

    return bp_fwhm, bl_fwhm


def fwhm_multibunch(Y_array, X_array, n_bunches,
                    bunch_spacing_buckets, bucket_size_tau,
                    bucket_tolerance=0.40, shift=0):
    """
    Computation of the bunch length and position from the FWHM
    assuming Gaussian line density for multibunch case.
    """

    bl_fwhm = np.zeros(n_bunches)
    bp_fwhm = np.zeros(n_bunches)

    if bm.device == 'GPU':
        X_array = X_array.get()
        Y_array = Y_array.get()
    for indexBunch in range(0, n_bunches):

        left_edge = indexBunch * bunch_spacing_buckets * bucket_size_tau -\
            bucket_tolerance * bucket_size_tau
        right_edge = indexBunch * bunch_spacing_buckets * bucket_size_tau +\
            bucket_size_tau + bucket_tolerance * bucket_size_tau
        indexes_bucket = np.where((X_array > left_edge) *
                                  (X_array < right_edge))[0]

        bp_fwhm[indexBunch], bl_fwhm[indexBunch] = fwhm(
            Y_array[indexes_bucket], X_array[indexes_bucket], shift)

    return bp_fwhm, bl_fwhm


def rms_multibunch(Y_array, X_array, n_bunches,
                   bunch_spacing_buckets, bucket_size_tau,
                   bucket_tolerance=0.40):
    """
    Computation of the rms bunch length (4sigma) and position.
    """

    bl_rms = np.zeros(n_bunches)
    bp_rms = np.zeros(n_bunches)

    if bm.device == 'GPU':
        X_array = X_array.get()
        Y_array = Y_array.get()

    for indexBunch in range(0, n_bunches):
        left_edge = indexBunch * bunch_spacing_buckets * bucket_size_tau -\
            bucket_tolerance * bucket_size_tau
        right_edge = indexBunch * bunch_spacing_buckets * bucket_size_tau +\
            bucket_size_tau + bucket_tolerance * bucket_size_tau

        indexes_bucket = np.where((X_array > left_edge) *
                                  (X_array < right_edge))[0]

        bp_rms[indexBunch], bl_rms[indexBunch] = rms(
            Y_array[indexes_bucket],
            X_array[indexes_bucket])

    return bp_rms, bl_rms




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to generate distributions**

:Authors: **Danilo Quartullo**, **Helga Timko**, **Alexandre Lasheen**,
          **Juan F. Esteban Mueller**, **Theodoros Argyropoulos**,
          **Joel Repond**
'''

from __future__ import absolute_import, division, print_function

import copy
import gc
import warnings
from builtins import range, str

import matplotlib.pyplot as plt
import numpy as np
import scipy
from packaging.version import Version

if Version(scipy.__version__) >= Version("1.14"):
    from scipy.integrate import cumulative_trapezoid as cumtrapz
else:
    from scipy.integrate import cumtrapz


from ..beam.profile import CutOptions, Profile
from ..trackers.utilities import (is_in_separatrix, minmax_location,
                                  potential_well_cut)
from ..utils import bmath as bm


def matched_from_line_density(beam, full_ring_and_RF, line_density_input=None,
                              main_harmonic_option='lowest_freq',
                              TotalInducedVoltage=None, plot=False,
                              figdir='fig', half_option='first',
                              extraVoltageDict=None, n_iterations=100,
                              n_points_potential=1e4, n_points_grid=int(1e3),
                              dt_margin_percent=0.40, n_points_abel=1e4,
                              bunch_length=None, line_density_type=None,
                              line_density_exponent=None, seed=None,
                              process_pot_well=True):
    '''
    *Function to generate a beam by inputing the line density. The distribution
    function is then reconstructed with the Abel transform and the particles
    randomly generated.*
    '''

    # Initialize variables depending on the accelerator parameters
    slippage_factor = full_ring_and_RF.RingAndRFSection_list[0].rf_params.eta_0[0]

    eom_factor_dE = abs(slippage_factor) / (2 * beam.beta**2. * beam.energy)
    eom_factor_potential = (np.sign(slippage_factor) * beam.Particle.charge /
                            (full_ring_and_RF.RingAndRFSection_list[0].rf_params.t_rev[0]))

    #: *Number of points to be used in the potential well calculation*
    n_points_potential = int(n_points_potential)
    # Generate potential well
    full_ring_and_RF.potential_well_generation(n_points=n_points_potential,
                                               dt_margin_percent=dt_margin_percent,
                                               main_harmonic_option=main_harmonic_option)
    potential_well = full_ring_and_RF.potential_well
    time_potential = full_ring_and_RF.potential_well_coordinates

    extra_potential = 0

    if extraVoltageDict is not None:
        extra_voltage_time_input = extraVoltageDict['time_array']
        extra_voltage_input = extraVoltageDict['voltage_array']
        extra_potential_input = - (eom_factor_potential *
                                   cumtrapz(extra_voltage_input,
                                            dx=extra_voltage_time_input[1] -
                                            extra_voltage_time_input[0], initial=0))
        extra_potential = np.interp(time_potential, extra_voltage_time_input,
                                    extra_potential_input)

    if line_density_type != 'user_input':
        # Time coordinates for the line density
        n_points_line_den = int(1e4)
        time_line_den = np.linspace(float(time_potential[0]), float(time_potential[-1]),
                                    n_points_line_den)
        line_den_resolution = time_line_den[1] - time_line_den[0]

        # Normalizing the line density
        line_density_ = line_density(time_line_den, line_density_type,
                                     bunch_length, exponent=line_density_exponent,
                                     bunch_position=(time_potential[0] + time_potential[-1]) / 2)

        line_density_ -= np.min(line_density_)
        line_density_ *= beam.n_macroparticles / np.sum(line_density_)

    elif line_density_type == 'user_input':
        # Time coordinates for the line density
        time_line_den = line_density_input['time_line_den']
        n_points_line_den = len(time_line_den)
        line_den_resolution = time_line_den[1] - time_line_den[0]

        # Normalizing the line density
        line_density_ = line_density_input['line_density']
        line_density_ -= np.min(line_density_)
        line_density_ *= beam.n_macroparticles / np.sum(line_density_)
    else:
        # GenerationError
        raise RuntimeError('The input for the matched_from_line_density ' +
                           'function was not recognized')

    induced_potential_final = 0

    if TotalInducedVoltage is not None:
        # Calculating the induced voltage
        induced_voltage_object = copy.deepcopy(TotalInducedVoltage)
        profile = induced_voltage_object.profile

        # Inputing new line density
        profile.cut_options.cut_left = time_line_den[0] - \
            0.5 * line_den_resolution
        profile.cut_options.cut_right = time_line_den[-1] + \
            0.5 * line_den_resolution
        profile.cut_options.n_slices = n_points_line_den
        profile.cut_options.cuts_unit = 's'
        profile.cut_options.set_cuts()
        profile.set_slices_parameters()
        profile.n_macroparticles = line_density_

        # Re-calculating the sources of wakes/impedances according to this
        # slicing
        induced_voltage_object.reprocess()

        # Calculating the induced voltage
        induced_voltage_object.induced_voltage_sum()
        induced_voltage = induced_voltage_object.induced_voltage

        # Calculating the induced potential
        induced_potential = -(eom_factor_potential * cumtrapz(induced_voltage,
                                                              dx=profile.bin_size, initial=0))

    # Centering the bunch in the potential well
    for i in range(0, n_iterations):
        if TotalInducedVoltage is not None:
            # Interpolating the potential well
            induced_potential_final = np.interp(time_potential,
                                                profile.bin_centers, induced_potential)

        # Induced voltage contribution
        total_potential = (potential_well + induced_potential_final +
                           extra_potential)

        # Potential well calculation around the separatrix
        if not process_pot_well:
            time_potential_sep, potential_well_sep = time_potential, total_potential
        else:
            time_potential_sep, potential_well_sep = potential_well_cut(
                time_potential, total_potential)

        minmax_positions_potential, minmax_values_potential = \
            minmax_location(time_potential_sep, potential_well_sep)
        minmax_positions_profile, minmax_values_profile = \
            minmax_location(time_line_den[line_density_ != 0],
                            line_density_[line_density_ != 0])

        n_minima_potential = len(minmax_positions_potential[0])
        n_maxima_profile = len(minmax_positions_profile[1])

        # Warnings
        if n_maxima_profile > 1:
            print('Warning: the profile has serveral max, the highest one ' +
                  'is taken. Be sure the profile is monotonous and not too noisy.')
            max_profile_pos = minmax_positions_profile[1][np.where(
                minmax_values_profile[1] == minmax_values_profile[1].max())]
        else:
            max_profile_pos = minmax_positions_profile[1]
        if n_minima_potential > 1:
            print('Warning: the potential well has serveral min, the deepest ' +
                  'one is taken. The induced potential is probably splitting ' +
                  'the potential well.')
            min_potential_pos = minmax_positions_potential[0][np.where(
                minmax_values_potential[0] == minmax_values_potential[0].min())]
        else:
            min_potential_pos = minmax_positions_potential[0]

        # Moving the bunch (not for the last iteration if intensity effects
        # are present)
        if TotalInducedVoltage is None:
            time_line_den -= max_profile_pos - min_potential_pos
            max_profile_pos -= max_profile_pos - min_potential_pos
        elif i != n_iterations - 1:
            time_line_den -= max_profile_pos - min_potential_pos
            # Update profile
            profile.cut_options.cut_left -= max_profile_pos - min_potential_pos
            profile.cut_options.cut_right -= max_profile_pos - min_potential_pos
            profile.cut_options.set_cuts()
            profile.set_slices_parameters()

    # Taking the first/second half of line density and potential
    n_points_abel = int(n_points_abel)

    abel_both_step = 1
    if half_option == 'both':
        abel_both_step = 2
        distribution_function_average = np.zeros((n_points_abel, 2))
        hamiltonian_average = np.zeros((n_points_abel, 2))

    for abel_index in range(0, abel_both_step):
        if half_option == 'first':
            half_indexes = np.where((time_line_den >= time_line_den[0]) *
                                    (time_line_den <= max_profile_pos))
        if half_option == 'second':
            half_indexes = np.where((time_line_den >= max_profile_pos) *
                                    (time_line_den <= time_line_den[-1]))
        if half_option == 'both' and abel_index == 0:
            half_indexes = np.where((time_line_den >= time_line_den[0]) *
                                    (time_line_den <= max_profile_pos))
        if half_option == 'both' and abel_index == 1:
            half_indexes = np.where((time_line_den >= max_profile_pos) *
                                    (time_line_den <= time_line_den[-1]))

        line_den_half = line_density_[half_indexes]
        time_half = time_line_den[half_indexes]
        potential_half = np.interp(time_half, time_potential_sep,
                                   potential_well_sep)
        potential_half = potential_half - np.min(potential_half)

        # Derivative of the line density
        line_den_diff = np.diff(line_den_half) / line_den_resolution

        time_line_den_diff = time_half[:-1] + line_den_resolution / 2
        line_den_diff = np.interp(time_half, time_line_den_diff, line_den_diff,
                                  left=0, right=0)

        # Interpolating the line density derivative and potential well for
        # Abel transform
        time_abel = np.linspace(
            float(time_half[0]), float(time_half[-1]), n_points_abel)
        line_den_diff_abel = np.interp(time_abel, time_half, line_den_diff)
        potential_abel = np.interp(time_abel, time_half, potential_half)

        distribution_function_ = np.zeros(n_points_abel)
        hamiltonian_coord = np.zeros(n_points_abel)

        # Abel transform
        warnings.filterwarnings("ignore")

        if (half_option == 'first') or (half_option == 'both' and
                                        abel_index == 0):
            for i in range(0, n_points_abel):
                integrand = (line_den_diff_abel[:i + 1] /
                             np.sqrt(potential_abel[:i + 1] - potential_abel[i]))

                if len(integrand) > 2:
                    integrand[-1] = integrand[-2] + (integrand[-2] -
                                                     integrand[-3])
                elif len(integrand) > 1:
                    integrand[-1] = integrand[-2]
                else:
                    integrand = np.array([0])

                distribution_function_[i] = (np.sqrt(eom_factor_dE) / np.pi *
                                             np.trapz(integrand, dx=line_den_resolution))

                hamiltonian_coord[i] = potential_abel[i]

        if (half_option == 'second') or (half_option == 'both' and
                                         abel_index == 1):
            for i in range(0, n_points_abel):
                integrand = (line_den_diff_abel[i:] /
                             np.sqrt(potential_abel[i:] - potential_abel[i]))

                if len(integrand) > 2:
                    integrand[0] = integrand[1] + (integrand[2] - integrand[1])
                if len(integrand) > 1:
                    integrand[0] = integrand[1]
                else:
                    integrand = np.array([0])

                distribution_function_[i] = -(np.sqrt(eom_factor_dE) / np.pi *
                                              np.trapz(integrand, dx=line_den_resolution))
                hamiltonian_coord[i] = potential_abel[i]

        warnings.filterwarnings("default")

        # Cleaning the distribution function from unphysical results
        distribution_function_[np.isnan(distribution_function_)] = 0
        distribution_function_[distribution_function_ < 0] = 0

        if half_option == 'both':
            hamiltonian_average[:, abel_index] = hamiltonian_coord
            distribution_function_average[:, abel_index] = \
                distribution_function_

    if half_option == 'both':
        hamiltonian_coord = hamiltonian_average[:, 0]
        distribution_function_ = (distribution_function_average[:, 0] +
                                  np.interp(hamiltonian_coord, hamiltonian_average[:, 1],
                                            distribution_function_average[:, 1])) / 2

    # Compute deltaE frame corresponding to the separatrix
    max_potential = np.max(potential_half)
    max_deltaE = np.sqrt(max_potential / eom_factor_dE)

    # Initializing the grids by reducing the resolution to a
    # n_points_grid*n_points_grid frame
    time_for_grid = np.linspace(float(time_line_den[0]), float(time_line_den[-1]),
                                n_points_grid)
    deltaE_for_grid = np.linspace(-float(max_deltaE),
                                  float(max_deltaE), n_points_grid)
    potential_well_for_grid = np.interp(time_for_grid, time_potential_sep,
                                        potential_well_sep)
    potential_well_for_grid = (potential_well_for_grid -
                               potential_well_for_grid.min())

    time_grid, deltaE_grid = np.meshgrid(time_for_grid, deltaE_for_grid)
    potential_well_grid = np.meshgrid(potential_well_for_grid,
                                      potential_well_for_grid)[0]

    hamiltonian_grid = eom_factor_dE * deltaE_grid**2 + potential_well_grid

    # Sort the distribution function and generate the density grid
    hamiltonian_argsort = np.argsort(hamiltonian_coord)
    hamiltonian_coord = hamiltonian_coord.take(hamiltonian_argsort)
    distribution_function_ = distribution_function_.take(hamiltonian_argsort)
    density_grid = np.interp(hamiltonian_grid, hamiltonian_coord,
                             distribution_function_)

    density_grid[np.isnan(density_grid)] = 0
    density_grid[density_grid < 0] = 0
    # Normalizing density
    density_grid = density_grid / np.sum(density_grid)
    reconstructed_line_den = np.sum(density_grid, axis=0)

    # Ploting the result
    if plot:
        plt.figure('Generated bunch')
        plt.plot(time_line_den, line_density_)
        plt.plot(time_for_grid, reconstructed_line_den /
                 np.max(reconstructed_line_den) * np.max(line_density_))
        plt.title('Line densities')
        if plot == 'show':
            plt.show()
        elif plot == 'savefig':
            fign = figdir + '/generated_bunch.png'
            plt.savefig(fign)

    # Populating the bunch
    populate_bunch(beam, time_grid, deltaE_grid, density_grid,
                   time_for_grid[1] - time_for_grid[0],
                   deltaE_for_grid[1] - deltaE_for_grid[0], seed)

    if TotalInducedVoltage is not None:
        # Inputing new line density
        profile.cut_options.cut_left = time_for_grid[0] - \
            0.5 * (time_for_grid[1] - time_for_grid[0])
        profile.cut_options.cut_right = time_for_grid[-1] + 0.5 * (
            time_for_grid[1] - time_for_grid[0])
        profile.cut_options.n_slices = n_points_grid
        profile.cut_options.set_cuts()
        profile.set_slices_parameters()
        profile.n_macroparticles = reconstructed_line_den * beam.n_macroparticles

        # Re-calculating the sources of wakes/impedances according to this
        # slicing
        induced_voltage_object.reprocess()

        # Calculating the induced voltage
        induced_voltage_object.induced_voltage_sum()
        gc.collect()
        return [hamiltonian_coord, distribution_function_], \
            induced_voltage_object

    gc.collect()
    return [hamiltonian_coord, distribution_function_],\
            [time_line_den, line_density_]


def matched_from_distribution_function(beam, full_ring_and_RF,
                                       distribution_function_input=None,
                                       distribution_user_table=None,
                                       main_harmonic_option='lowest_freq',
                                       TotalInducedVoltage=None,
                                       n_iterations=1, n_points_potential=1e4,
                                       n_points_grid=int(1e3),
                                       dt_margin_percent=0.40,
                                       extraVoltageDict=None, seed=None,
                                       distribution_exponent=None,
                                       distribution_type=None,
                                       emittance=None, bunch_length=None,
                                       bunch_length_fit=None,
                                       distribution_variable='Hamiltonian',
                                       process_pot_well=True,
                                       turn_number=0):
    '''
    *Function to generate a beam by inputing the distribution function (by
    choosing the type of distribution and the emittance).
    The potential well is preprocessed to check for the min/max and center
    the frame around the separatrix.
    An error will be raised if there is not a full potential well (2 max
    and 1 min at least), or if there are several wells (more than 2 max and
    1 min, this case will be treated in the future).
    An adjustable margin (40% by default) is applied in order to be able to
    catch the min/max of the potential well that might be on the edge of the
    frame. The slippage factor should be updated to take the higher orders.
    Outputs should be added in order for the user to check step by step if
    his bunch is going to be well generated. More detailed 'step by step'
    documentation should be implemented
    The user can input a custom distribution function by setting the parameter
    distribution_type = 'user_input' and passing the function in the
    parameter distribution_options['function'], with the following definition:
    distribution_function(action_array, dist_type, length, exponent=None).
    The user can also add an input table by setting the parameter
    distribution_type = 'user_input_table',
    distribution_options['user_table_action'] = array of action (in H or in J)
    and distribution_options['user_table_distribution']*
    '''

    # Loading the distribution function if provided by the user
    if distribution_function_input is not None:
        distribution_function_ = distribution_function_input
    else:
        distribution_function_ = distribution_function

    # Initialize variables depending on the accelerator parameters
    slippage_factor = full_ring_and_RF.RingAndRFSection_list[0].rf_params.eta_0[turn_number]
    beta = full_ring_and_RF.RingAndRFSection_list[0].rf_params.beta[turn_number]
    energy = full_ring_and_RF.RingAndRFSection_list[0].rf_params.energy[turn_number]

    eom_factor_dE = abs(slippage_factor) / (2 * beta**2. * energy)
    eom_factor_potential = (np.sign(slippage_factor) * beam.Particle.charge /
                            (full_ring_and_RF.RingAndRFSection_list[0].rf_params.t_rev[turn_number]))

    #: *Number of points to be used in the potential well calculation*
    n_points_potential = int(n_points_potential)
    # Generate potential well
    full_ring_and_RF.potential_well_generation(turn=turn_number,
                                               n_points=n_points_potential,
                                               dt_margin_percent=dt_margin_percent,
                                               main_harmonic_option=main_harmonic_option)
    potential_well = full_ring_and_RF.potential_well
    time_potential = full_ring_and_RF.potential_well_coordinates

    induced_potential = 0

    # Extra potential from previous bunches (for multi-bunch generation)
    extra_potential = 0
    if extraVoltageDict is not None:
        extra_voltage_time_input = extraVoltageDict['time_array']
        extra_voltage_input = extraVoltageDict['voltage_array']
        extra_potential_input = -(eom_factor_potential *
                                  cumtrapz(extra_voltage_input, dx=extra_voltage_time_input[1] -
                                           extra_voltage_time_input[0], initial=0))
        extra_potential = np.interp(time_potential, extra_voltage_time_input,
                                    extra_potential_input)

    total_potential = potential_well + induced_potential + extra_potential

    if not TotalInducedVoltage:
        n_iterations = 1
    else:
        induced_voltage_object = copy.deepcopy(TotalInducedVoltage)
        profile = induced_voltage_object.profile

    dE_trajectory = np.zeros(n_points_potential)
    for i in range(n_iterations):
        old_potential = copy.deepcopy(total_potential)

        # Adding the induced potential to the RF potential
        total_potential = (potential_well + induced_potential +
                           extra_potential)

        sse = np.sqrt(np.sum((old_potential - total_potential)**2))

        print('Matching the bunch... (iteration: ' + str(i) + ' and sse: ' +
              str(sse) + ')')

        # Process the potential well in order to take a frame around the separatrix
        if not process_pot_well:
            time_potential_sep, potential_well_sep = time_potential, total_potential
        else:
            time_potential_sep, potential_well_sep = potential_well_cut(
                time_potential, total_potential)

        # Potential is shifted to put the minimum on 0
        potential_well_sep = potential_well_sep - np.min(potential_well_sep)

        # Compute deltaE frame corresponding to the separatrix
        max_potential = np.max(potential_well_sep)
        max_deltaE = np.sqrt(max_potential / eom_factor_dE)

        # Initializing the grids by reducing the resolution to a
        # n_points_grid*n_points_grid frame
        time_potential_low_res = np.linspace(float(time_potential_sep[0]),
                                             float(time_potential_sep[-1]),
                                             n_points_grid)
        time_resolution_low = (time_potential_low_res[1] -
                               time_potential_low_res[0])
        deltaE_coord_array = np.linspace(-float(max_deltaE), float(max_deltaE),
                                         n_points_grid)
        potential_well_low_res = np.interp(time_potential_low_res,
                                           time_potential_sep, potential_well_sep)
        time_grid, deltaE_grid = np.meshgrid(time_potential_low_res,
                                             deltaE_coord_array)
        potential_well_grid = np.meshgrid(potential_well_low_res,
                                          potential_well_low_res)[0]

        # Computing the action J by integrating the dE trajectories
        J_array_dE0 = np.zeros(n_points_grid)

        full_ring_and_RF2 = copy.deepcopy(full_ring_and_RF)
        for j in range(n_points_grid):
            # Find left and right time coordinates for a given hamiltonian
            # value
            time_indexes = np.where(potential_well_low_res <=
                                    potential_well_low_res[j])[0]
            left_time = time_potential_low_res[np.max((0, time_indexes[0]))]
            right_time = time_potential_low_res[np.min((time_indexes[-1],
                                                        n_points_grid - 1))]
            # Potential well calculation with high resolution in that frame
            time_potential_high_res = np.linspace(float(left_time), float(right_time),
                                                  n_points_potential)
            full_ring_and_RF2.potential_well_generation(
                n_points=n_points_potential,
                time_array=time_potential_high_res,
                main_harmonic_option=main_harmonic_option)
            pot_well_high_res = full_ring_and_RF2.potential_well

            if TotalInducedVoltage is not None and i != 0:
                induced_potential_hires = np.interp(time_potential_high_res,
                                                    time_potential, induced_potential +
                                                    extra_potential, left=0, right=0)
                pot_well_high_res += induced_potential_hires
                pot_well_high_res -= pot_well_high_res.min()

            # Integration to calculate action
            dE_trajectory[pot_well_high_res <= potential_well_low_res[j]] = \
                np.sqrt((potential_well_low_res[j] -
                         pot_well_high_res[pot_well_high_res <=
                                           potential_well_low_res[j]]) / eom_factor_dE)
            dE_trajectory[pot_well_high_res > potential_well_low_res[j]] = 0

            J_array_dE0[j] = 1 / np.pi * np.trapz(dE_trajectory,
                                                  dx=time_potential_high_res[1] - time_potential_high_res[0])

        # Sorting the H and J functions to be able to interpolate J(H)
        H_array_dE0 = potential_well_low_res
        sorted_H_dE0 = H_array_dE0[H_array_dE0.argsort()]
        sorted_J_dE0 = J_array_dE0[H_array_dE0.argsort()]

        # Calculating the H and J grid
        H_grid = eom_factor_dE * deltaE_grid**2 + potential_well_grid
        J_grid = np.interp(H_grid, sorted_H_dE0, sorted_J_dE0, left=0,
                           right=np.inf)

        # Choice of either H or J as the variable used
        if distribution_variable == 'Action':
            sorted_X_dE0 = sorted_J_dE0
            X_grid = J_grid
        elif distribution_variable == 'Hamiltonian':
            sorted_X_dE0 = sorted_H_dE0
            X_grid = H_grid
        else:
            # DistributionError
            raise RuntimeError('The distribution_variable option was not ' +
                               'recognized')

        # Computing bunch length as a function of H/J if needed
        # Bunch length can be calculated as 4-rms, Gaussian fit, or FWHM
        if bunch_length is not None:
            X0 = X0_from_bunch_length(bunch_length, bunch_length_fit,
                                      X_grid, sorted_X_dE0, n_points_grid,
                                      time_potential_low_res, distribution_function_,
                                      distribution_type, distribution_exponent, beam,
                                      full_ring_and_RF)

        elif emittance is not None:
            if distribution_variable == 'Action':
                X0 = emittance / (2 * np.pi)
            elif distribution_variable == 'Hamiltonian':
                X0 = np.interp(emittance / (2 * np.pi), sorted_J_dE0,
                               sorted_H_dE0)

        # Computing the density grid
        if distribution_user_table is None:
            density_grid = distribution_function_(X_grid, distribution_type,
                                                  X0, distribution_exponent)
        else:
            density_grid = np.interp(X_grid,
                                     distribution_user_table['user_table_action'],
                                     distribution_user_table['user_table_distribution'])

        # Normalizing the grid
        density_grid[H_grid > np.max(H_array_dE0)] = 0
        density_grid = density_grid / np.sum(density_grid)

        # Calculating the line density
        line_density_ = np.sum(density_grid, axis=0)
        line_density_ *= beam.n_macroparticles / np.sum(line_density_)

        # Induced voltage contribution
        if TotalInducedVoltage is not None:
            # Inputing new line density
            profile.cut_options.cut_left = time_potential_low_res[0] - \
                0.5 * time_resolution_low
            profile.cut_options.cut_right = time_potential_low_res[-1] + \
                0.5 * time_resolution_low
            profile.cut_options.n_slices = n_points_grid
            profile.cut_options.cuts_unit = 's'
            profile.cut_options.set_cuts()
            profile.set_slices_parameters()
            profile.n_macroparticles = line_density_

            # Re-calculating the sources of wakes/impedances according to this
            # slicing
            induced_voltage_object.reprocess()

            # Calculating the induced voltage
            induced_voltage_object.induced_voltage_sum()
            induced_voltage = induced_voltage_object.induced_voltage

            # Calculating the induced potential
            induced_potential_low_res = -(eom_factor_potential *
                                          cumtrapz(induced_voltage,
                                                   dx=time_resolution_low,
                                                   initial=0))
            induced_potential = np.interp(time_potential,
                                          time_potential_low_res, induced_potential_low_res,
                                          left=0, right=0)
        del full_ring_and_RF2
        gc.collect()
    # Populating the bunch
    populate_bunch(beam, time_grid, deltaE_grid, density_grid,
                   time_resolution_low, deltaE_coord_array[1] -
                   deltaE_coord_array[0], seed)

    if TotalInducedVoltage is not None:
        return [time_potential_low_res, line_density_], induced_voltage_object
    else:
        return [time_potential_low_res, line_density_]


def X0_from_bunch_length(bunch_length, bunch_length_fit, X_grid, sorted_X_dE0,
                         n_points_grid, time_potential_low_res,
                         distribution_function_, distribution_type,
                         distribution_exponent, beam, full_ring_and_RF):
    '''
    Function to find the corresponding H0 or J0 for a given bunch length.
    Used by matched_from_distribution_function()
    '''
    tau = 0.0

    # Initial values for iteration
    X_low = sorted_X_dE0[0]
    X_hi = sorted_X_dE0[-1]
    X_min = sorted_X_dE0[0]
    X_max = sorted_X_dE0[-1]
    X_accuracy = (sorted_X_dE0[1] - sorted_X_dE0[0]) / 2.0

    bin_size = (time_potential_low_res[1] - time_potential_low_res[0])

    # Iteration to find H0/J0 from the bunch length
    while np.abs(bunch_length - tau) > bin_size:
        # Takes middle point of the interval [X_low,X_hi]
        X0 = 0.5 * (X_low + X_hi)

        if bunch_length_fit == 'full':
            bunchIndices = np.where(np.sum(X_grid <= X0, axis=0))[0]
            tau = (time_potential_low_res[bunchIndices][-1] -
                   time_potential_low_res[bunchIndices][0])
        else:
            # Calculating the line density for the parameter X0
            density_grid = distribution_function_(X_grid,
                                                  distribution_type, X0, distribution_exponent)

            density_grid = density_grid / np.sum(density_grid)
            line_density_ = np.sum(density_grid, axis=0)

            # Calculating the bunch length of that line density
            if (line_density_ > 0).any():
                tau = 4.0 * np.sqrt(np.sum((time_potential_low_res -
                                            np.sum(line_density_ * time_potential_low_res) /
                                            np.sum(line_density_))**2 * line_density_) /
                                    np.sum(line_density_))

                if bunch_length_fit is not None:
                    profile = Profile(
                        beam, CutOptions=CutOptions(cut_left=time_potential_low_res[0] -
                                                    0.5 * bin_size, cut_right=time_potential_low_res[-1] +
                                                    0.5 * bin_size, n_slices=n_points_grid, RFSectionParameters=full_ring_and_RF.RingAndRFSection_list[0].rf_params))
#                     profile = Profile(
#                       full_ring_and_RF.RingAndRFSection_list[0].rf_params,
#                       beam, n_points_grid, cut_left=time_potential_low_res[0] -
#                       0.5*bin_size , cut_right=time_potential_low_res[-1] +
#                       0.5*bin_size)

                    profile.n_macroparticles = line_density_

                    if bunch_length_fit == 'gauss':
                        profile.bl_gauss = tau
                        profile.bp_gauss = np.sum(line_density_ *
                                                  time_potential_low_res) / np.sum(line_density_)
                        profile.gaussian_fit()
                        tau = profile.bl_gauss
                    elif bunch_length_fit == 'fwhm':
                        profile.fwhm()
                        tau = profile.bunchLength

        # Update of the interval for the next iteration
        if tau >= bunch_length:
            X_hi = X0
        else:
            X_low = X0

        if (X_max - X0) < X_accuracy:
            print('WARNING: The bucket is too small to have the ' +
                  'desired bunch length! Input is %.2e, ' % (bunch_length) +
                  'the generation gave %.2e, ' % (tau) +
                  'the error is %.2e' % (bunch_length - tau))
            break

        if (X0 - X_min) < X_accuracy:
            print('WARNING: The desired bunch length is too small ' +
                  'to be generated accurately!')

#    return 0.5 * (X_low + X_hi)
    return X0


def populate_bunch(beam, time_grid, deltaE_grid, density_grid, time_step,
                   deltaE_step, seed):
    '''
    *Method to populate the bunch using a random number generator from the
    particle density in phase space.*
    '''
    # Initialise the random number generator
    np.random.seed(seed=seed)
    # Generating particles randomly inside the grid cells according to the
    # provided density_grid
    indexes = np.random.choice(np.arange(0, np.size(density_grid)),
                               beam.n_macroparticles, p=density_grid.flatten())

    # Randomize particles inside each grid cell (uniform distribution)
    beam.dt = (np.ascontiguousarray(time_grid.flatten()[indexes] +
                                    (np.random.rand(beam.n_macroparticles) - 0.5) * time_step)).astype(dtype=bm.precision.real_t, order='C', copy=False)
    beam.dE = (np.ascontiguousarray(deltaE_grid.flatten()[indexes] +
                                    (np.random.rand(beam.n_macroparticles) - 0.5) * deltaE_step)).astype(dtype=bm.precision.real_t, order='C', copy=False)


def distribution_function(action_array, dist_type, length, exponent=None):
    '''
    *Distribution function (formulas from Laclare).*
    '''

    if dist_type in ['binomial', 'waterbag', 'parabolic_amplitude',
                     'parabolic_line']:
        if dist_type == 'waterbag':
            exponent = 0
        elif dist_type == 'parabolic_amplitude':
            exponent = 1
        elif dist_type == 'parabolic_line':
            exponent = 0.5

        warnings.filterwarnings("ignore")
        distribution_function_ = (1 - action_array / length)**exponent
        warnings.filterwarnings("default")
        distribution_function_[action_array > length] = 0
    elif dist_type == 'gaussian':
        distribution_function_ = np.exp(- 2 * action_array / length)
    else:
        # DistributionError
        raise RuntimeError('The dist_type option was not recognized')

    return distribution_function_


def line_density(coord_array, dist_type, bunch_length, bunch_position=0,
                 exponent=None):
    '''
    *Line density*
    '''

    if dist_type in ['binomial', 'waterbag', 'parabolic_amplitude',
                     'parabolic_line']:
        if dist_type == 'waterbag':
            exponent = 0
        elif dist_type == 'parabolic_amplitude':
            exponent = 1
        elif dist_type == 'parabolic_line':
            exponent = 0.5

        warnings.filterwarnings("ignore")
        line_density_ = ((1 - (2.0 * (coord_array - bunch_position) /
                         bunch_length)**2)**(exponent + 0.5))
        warnings.filterwarnings("default")
        line_density_[np.abs(coord_array - bunch_position)
                      > bunch_length / 2] = 0

    elif dist_type == 'gaussian':
        sigma = bunch_length / 4
        line_density_ = np.exp(-(coord_array - bunch_position)
                               ** 2 / (2 * sigma**2))

    elif dist_type == 'cosine_squared':
        warnings.filterwarnings("ignore")
        line_density_ = (np.cos(np.pi * (coord_array - bunch_position) /
                                bunch_length)**2)
        warnings.filterwarnings("default")
        line_density_[np.abs(coord_array - bunch_position)
                      > bunch_length / 2] = 0
    else:
        # DistributionError
        raise RuntimeError('The dist_type option was not recognized')

    return line_density_


def bigaussian(Ring, RFStation, Beam, sigma_dt, sigma_dE=None, seed=1234,
               reinsertion=False):
    r"""Function generating a Gaussian beam both in time and energy
    coordinates. Fills Beam.dt and Beam.dE arrays.

    Parameters
    ----------
    Ring : class
        A Ring type class
    RFStation : class
        An RFStation type class
    Beam : class
        A Beam type class
    sigma_dt : float
        R.m.s. extension of the Gaussian in time
    sigma_dE : float (optional)
        R.m.s. extension of the Gaussian in energy; default is None and will
        match the energy coordinate according to bucket height and sigma_dt
    seed : int (optional)
        Fixed seed to have a reproducible distribution
    reinsertion : bool (optional)
        Re-insert particles that are generated outside the separatrix into the
        bucket; default in False

    """
    
    if sigma_dE is None:
        sigma_dE = _get_dE_from_dt(Ring, RFStation, sigma_dt)
    counter = RFStation.counter[0]
    omega_rf = RFStation.omega_rf[0, counter]
    phi_s = RFStation.phi_s[counter]
    phi_rf = RFStation.phi_rf[0, counter]
    eta0 = RFStation.eta_0[counter]
    
    # RF wave is shifted by Pi below transition
    if eta0 < 0:
        phi_rf -= np.pi

    Beam.sigma_dt = sigma_dt
    Beam.sigma_dE = sigma_dE

    # Generate coordinates. For reproducibility, a separate random number stream is used for dt and dE
    rng_dt = np.random.default_rng(seed)
    rng_dE = np.random.default_rng(seed + 1)

    Beam.dt = sigma_dt * rng_dt.normal(size=Beam.n_macroparticles).astype(dtype=bm.precision.real_t, order='C', copy=False) + \
        (phi_s - phi_rf) / omega_rf
    Beam.dE = sigma_dE * \
        rng_dE.normal(size=Beam.n_macroparticles).astype(
            dtype=bm.precision.real_t, order='C')

    # Re-insert if necessary
    if reinsertion:

        itemindex = bm.where(is_in_separatrix(Ring, RFStation, Beam,
                                              Beam.dt, Beam.dE) == False)[0]
        while itemindex.size > 0:

            Beam.dt[itemindex] = sigma_dt * rng_dt.normal(size=itemindex.size).astype(dtype=bm.precision.real_t, order='C', copy=False) \
                + (phi_s - phi_rf) / omega_rf

            Beam.dE[itemindex] = sigma_dE * rng_dE.normal(
                size=itemindex.size).astype(dtype=bm.precision.real_t, order='C')

            itemindex = bm.where(is_in_separatrix(Ring, RFStation, Beam,
                                                  Beam.dt, Beam.dE) == False)[0]


def parabolic(Ring, RFStation, Beam,
              bunch_length,
              bunch_position=None, 
              bunch_energy=None,
              energy_spread=None,
              seed=1234):
    r"""Generate a bunch of particles distributed as a parabola in phase space.

    Parameters
    ----------
    Ring : class
        A Ring type class
    RFStation : class
        An RFStation type class
    Beam : class
        A Beam type class
    bunch_length : float
        The length in time [s] of the bunch
    bunch_position : float (optional)
        The position in time [s] of the center of mass of the bunch
    bunch_energy : float (optional)
        The position in energy [eV] of the center of mass of the bunch
        (relative to the synchronous energy)
    energy_spread : float (optional)
        The spread in energy [eV] of the bunch
    seed : int (optional)
        Fixed seed to have a reproducible distribution

    """
    
    # Getting the position and spread if not defined by user
    counter = RFStation.counter[0]
    omega_rf = RFStation.omega_rf[0, counter]
    phi_s = RFStation.phi_s[counter]
    phi_rf = RFStation.phi_rf[0, counter]
    eta0 = RFStation.eta_0[counter]
    
    # RF wave is shifted by Pi below transition
    if eta0 < 0:
        phi_rf -= np.pi
    
    if bunch_position is None:
        bunch_position = (phi_s - phi_rf) / omega_rf
        
    if bunch_energy is None:
        bunch_energy = 0
    
    if energy_spread is None:
        energy_spread = _get_dE_from_dt(Ring, RFStation, bunch_length)

    # Generating time and energy arrays
    time_array = np.linspace(bunch_position - bunch_length / 2,
                             bunch_position + bunch_length / 2,
                             100)
    energy_array = np.linspace(bunch_energy - energy_spread / 2,
                               bunch_energy + energy_spread / 2,
                               100)

    # Getting Hamiltonian on a grid
    dt_grid, deltaE_grid = np.meshgrid(
        time_array, energy_array)

    # Bin sizes
    bin_dt = time_array[1] - time_array[0]
    bin_energy = energy_array[1] - energy_array[0]

    # Density grid
    isodensity_lines = ((dt_grid - bunch_position) / bunch_length * 2)**2. + \
        ((deltaE_grid - bunch_energy) / energy_spread * 2)**2.
    density_grid = 1 - isodensity_lines**2.
    density_grid[density_grid < 0] = 0
    density_grid /= np.sum(density_grid)
    
    populate_bunch(Beam, dt_grid, deltaE_grid, density_grid, bin_dt,
                   bin_energy, seed)


def _get_dE_from_dt(Ring, RFStation, dt_amplitude):
    r"""A routine to evaluate the dE amplitude from dt following a single
    RF Hamiltonian.

    Parameters
    ----------
    Ring : class
        A Ring type class
    RFStation : class
        An RFStation type class
    dt_amplitude : float
        Full amplitude of the particle oscillation in [s]

    Returns
    -------
    dE_amplitude : float
        Full amplitude of the particle oscillation in [eV]

    """

    warnings.filterwarnings("once")
    if Ring.n_sections > 1:
        warnings.warn("WARNING in bigaussian(): the usage of several" +
                        " sections is not yet implemented. Ignoring" +
                        " all but the first!")
    if RFStation.n_rf > 1:
        warnings.warn("WARNING in bigaussian(): the usage of multiple RF" +
                        " systems is not yet implemented. Ignoring" +
                        " higher harmonics!")

    counter = RFStation.counter[0]

    harmonic = RFStation.harmonic[0, counter]
    energy = RFStation.energy[counter]
    beta = RFStation.beta[counter]
    omega_rf = RFStation.omega_rf[0, counter]
    phi_s = RFStation.phi_s[counter]
    phi_rf = RFStation.phi_rf[0, counter]
    eta0 = RFStation.eta_0[counter]

    # RF wave is shifted by Pi below transition
    if eta0 < 0:
        phi_rf -= np.pi

    # Calculate dE_amplitude from dt_amplitude using single-harmonic Hamiltonian
    voltage = RFStation.charge * \
        RFStation.voltage[0, counter]
    eta0 = RFStation.eta_0[counter]

    phi_b = omega_rf * dt_amplitude + phi_s
    dE_amplitude = np.sqrt(voltage * energy * beta**2
                        * (np.cos(phi_b) - np.cos(phi_s) + (phi_b - phi_s) * np.sin(phi_s))
                        / (np.pi * harmonic * np.fabs(eta0)))
    
    return dE_amplitude



'''
BLonD math wrapper functions

@author Konstantinos Iliakis
@date 20.10.2017
'''

import ctypes as ct
import os
import sys
import numpy as np

# from .. import get_libblond
from . import c_complex64, c_complex128, c_real, precision


__LIBBLOND = None


def load_libblond(precision='single'):
    '''Locates and initializes the blond compiled library
    @param precision: The floating point precision of the calculations. Can be 'single' or 'double'.
    '''
    global __LIBBLOND
    libblond_path = os.environ.get('LIBBLOND', None)
    path = os.path.realpath(__file__)
    basepath = os.sep.join(path.split(os.sep)[:-1])
    # print(f'Loading the BLonD library for {precision} precision...')
    try:
        if 'posix' in os.name:
            if libblond_path:
                libblond_path = os.path.abspath(libblond_path)
            else:
                libblond_path = os.path.join(
                    basepath, f'../cpp_routines/libblond_{precision}.so')
            __LIBBLOND = ct.CDLL(libblond_path)
        elif 'win' in sys.platform:
            if libblond_path:
                libblond_path = os.path.abspath(libblond_path)
            else:
                libblond_path = os.path.join(
                    basepath, f'../cpp_routines/libblond_{precision}.dll')
            if hasattr(os, 'add_dll_directory'):
                os.add_dll_directory(os.path.dirname(libblond_path))
                __LIBBLOND = ct.CDLL(libblond_path, winmode=0)
            else:
                __LIBBLOND = ct.CDLL(libblond_path)
        else:
            print('YOU DO NOT HAVE A WINDOWS OR UNIX OPERATING SYSTEM. ABORTING.')
            sys.exit()
    except OSError:
        # Silently pass. An alternative backend can be used.
        pass


load_libblond(precision='double')


def get_libblond():
    '''Returns the blond library.
    '''
    return __LIBBLOND


def __getPointer(x):
    return x.ctypes.data_as(ct.c_void_p)


def __getLen(x):
    return ct.c_int(len(x))


# Similar to np.where with a condition of more_than < x < less_than
# You need to define at least one of more_than, less_than
# @return: a bool array, size equal to the input,
#           True: element satisfied the cond, False: otherwise
def where_cpp(x, more_than=None, less_than=None, result=None):
    if result is None:
        result = np.empty_like(x, dtype=bool)
    if more_than is None and less_than is not None:
        get_libblond().where_less_than(__getPointer(x), x.size,
                                       ct.c_double(less_than),
                                       __getPointer(result))
    elif more_than is not None and less_than is None:
        get_libblond().where_more_than(__getPointer(x), x.size,
                                       ct.c_double(more_than),
                                       __getPointer(result))

    elif more_than is not None and less_than is not None:
        get_libblond().where_more_less_than(__getPointer(x), x.size,
                                            ct.c_double(more_than),
                                            ct.c_double(less_than),
                                            __getPointer(result))

    else:
        raise RuntimeError(
            '[bmath:where] You need to define at least one of more_than, less_than')
    return result


def add_cpp(a, b, result=None, inplace=False):
    if (len(a) != len(b)):
        raise ValueError(
            'operands could not be broadcast together with shapes ',
            a.shape, b.shape)
    if a.dtype != b.dtype:
        raise TypeError(
            'given arrays not of the same type ', a.dtype(), b.dtype())

    if (result is None) and (not inplace):
        result = np.empty_like(a, order='C')

    if (a.dtype == 'int32'):
        if inplace:
            get_libblond().add_int_vector_inplace(__getPointer(a), __getPointer(b),
                                                  __getLen(a))
        else:
            get_libblond().add_int_vector(__getPointer(a), __getPointer(b),
                                          __getLen(a), __getPointer(result))
    elif (a.dtype == 'int64'):
        if inplace:
            get_libblond().add_longint_vector_inplace(__getPointer(a), __getPointer(b),
                                                      __getLen(a))
        else:
            get_libblond().add_longint_vector(__getPointer(a), __getPointer(b),
                                              __getLen(a), __getPointer(result))

    elif (a.dtype == 'float64'):
        if inplace:
            get_libblond().add_double_vector_inplace(__getPointer(a), __getPointer(b),
                                                     __getLen(a))
        else:
            get_libblond().add_double_vector(__getPointer(a), __getPointer(b),
                                             __getLen(a), __getPointer(result))
    elif (a.dtype == 'float32'):
        if inplace:
            get_libblond().add_float_vector_inplace(__getPointer(a), __getPointer(b),
                                                    __getLen(a))
        else:
            get_libblond().add_float_vector(__getPointer(a), __getPointer(b),
                                            __getLen(a), __getPointer(result))

    elif (a.dtype == 'uint16'):
        if inplace:
            get_libblond().add_uint16_vector_inplace(__getPointer(a), __getPointer(b),
                                                     __getLen(a))
        else:
            get_libblond().add_uint16_vector(__getPointer(a), __getPointer(b),
                                             __getLen(a), __getPointer(result))
    elif (a.dtype == 'uint32'):
        if inplace:
            get_libblond().add_uint32_vector_inplace(__getPointer(a), __getPointer(b),
                                                     __getLen(a))
        else:
            get_libblond().add_uint32_vector(__getPointer(a), __getPointer(b),
                                             __getLen(a), __getPointer(result))

    else:
        raise TypeError('type ', a.dtype, ' is not supported')

    return result


def mul_cpp(a, b, result=None):
    if (type(a) == np.ndarray and type(b) != np.ndarray):
        if result is None:
            result = np.empty_like(a, order='C')

        if (a.dtype == 'int32'):
            get_libblond().scalar_mul_int32(__getPointer(a), ct.c_int32(np.int32(b)),
                                            __getLen(a), __getPointer(result))
        elif (a.dtype == 'int64'):
            get_libblond().scalar_mul_int64(__getPointer(a), ct.c_int64(np.int64(b)),
                                            __getLen(a), __getPointer(result))
        elif (a.dtype == 'float32'):
            get_libblond().scalar_mul_float32(__getPointer(a), ct.c_float(np.float32(b)),
                                              __getLen(a), __getPointer(result))
        elif (a.dtype == 'float64'):
            get_libblond().scalar_mul_float64(__getPointer(a), ct.c_double(np.float64(b)),
                                              __getLen(a), __getPointer(result))
        elif (a.dtype == 'complex64'):
            get_libblond().scalar_mul_compex64(__getPointer(a), c_complex64(np.complex64(b)),
                                               __getLen(a), __getPointer(result))
        elif (a.dtype == 'complex128'):
            get_libblond().scalar_mul_complex128(__getPointer(a), c_complex128(np.complex128(b)),
                                                 __getLen(a), __getPointer(result))
        else:
            raise TypeError('type ', a.dtype, ' is not supported')

    elif (type(b) == np.ndarray and type(a) != np.ndarray):
        return mul_cpp(b, a, result)
    elif (type(a) == np.ndarray and type(b) == np.ndarray):
        if result is None:
            result = np.empty_like(a, order='C')

        if (a.dtype == 'int32'):
            get_libblond().vector_mul_int32(__getPointer(a), __getPointer(b),
                                            __getLen(a), __getPointer(result))
        elif (a.dtype == 'int64'):
            get_libblond().vector_mul_int64(__getPointer(a), __getPointer(b),
                                            __getLen(a), __getPointer(result))
        elif (a.dtype == 'float32'):
            get_libblond().vector_mul_float32(__getPointer(a), __getPointer(b),
                                              __getLen(a), __getPointer(result))
        elif (a.dtype == 'float64'):
            get_libblond().vector_mul_float64(__getPointer(a), __getPointer(b),
                                              __getLen(a), __getPointer(result))
        elif (a.dtype == 'complex64'):
            get_libblond().vector_mul_compex64(__getPointer(a), __getPointer(b),
                                               __getLen(a), __getPointer(result))
        elif (a.dtype == 'complex128'):
            get_libblond().vector_mul_complex128(__getPointer(a), __getPointer(b),
                                                 __getLen(a), __getPointer(result))
        else:
            raise TypeError('type ', a.dtype, ' is not supported')
    else:
        raise TypeError(
            'types {} and {} are not supported'.format(type(a), type(b)))
    return result


def argmin_cpp(x):
    get_libblond().min_idx.restype = ct.c_int
    return get_libblond().min_idx(__getPointer(x), __getLen(x))


def argmax_cpp(x):
    get_libblond().max_idx.restype = ct.c_int
    return get_libblond().max_idx(__getPointer(x), __getLen(x))


def linspace_cpp(start, stop, num=50, retstep=False, result=None):
    if result is None:
        result = np.empty(num, dtype=float)
    get_libblond().linspace(c_real(start), c_real(stop),
                            ct.c_int(num), __getPointer(result))
    if retstep:
        return result, 1. * (stop - start) / (num - 1)
    else:
        return result


def arange_cpp(start, stop, step, dtype=float, result=None):
    size = int(np.ceil((stop - start) / step))
    if result is None:
        result = np.empty(size, dtype=dtype)
    if dtype == float:
        get_libblond().arange_double(c_real(start), c_real(stop),
                                     c_real(step), __getPointer(result))
    elif dtype == int:
        get_libblond().arange_int(ct.c_int(start), ct.c_int(stop),
                                  ct.c_int(step), __getPointer(result))

    return result


def sum_cpp(x):
    get_libblond().sum.restype = ct.c_double
    return get_libblond().sum(__getPointer(x), __getLen(x))


def sort_cpp(x, reverse=False):
    if x.dtype == 'int32':
        get_libblond().sort_int(__getPointer(x), __getLen(x), ct.c_bool(reverse))
    elif x.dtype == 'float64':
        get_libblond().sort_double(__getPointer(x), __getLen(x), ct.c_bool(reverse))
    elif x.dtype == 'int64':
        get_libblond().sort_longint(__getPointer(x), __getLen(x), ct.c_bool(reverse))
    else:
        # SortError
        raise RuntimeError('[sort] Datatype %s not supported' % x.dtype)
    return x


def convolve(signal, kernel, mode='full', result=None):
    if mode != 'full':
        # ConvolutionError
        raise RuntimeError('[convolve] Only full mode is supported')
    if result is None:
        result = np.empty(len(signal) + len(kernel) - 1, dtype=float)
    get_libblond().convolution(__getPointer(signal), __getLen(signal),
                               __getPointer(kernel), __getLen(kernel),
                               __getPointer(result))
    return result


def mean_cpp(x):
    if isinstance(x[0], np.float32):
        get_libblond().meanf.restype = ct.c_float
        return get_libblond().meanf(__getPointer(x), __getLen(x))
    elif isinstance(x[0], np.float64):
        get_libblond().mean.restype = ct.c_double
        return get_libblond().mean(__getPointer(x), __getLen(x))


def std_cpp(x):
    if isinstance(x[0], np.float32):
        get_libblond().stdevf.restype = ct.c_float
        return get_libblond().stdevf(__getPointer(x), __getLen(x))
    elif isinstance(x[0], np.float64):
        get_libblond().stdev.restype = ct.c_double
        return get_libblond().stdev(__getPointer(x), __getLen(x))


def sin_cpp(x, result=None):
    if isinstance(x, np.ndarray) and isinstance(x[0], np.float64):
        if result is None:
            result = np.empty(len(x), dtype=np.float64, order='C')
        get_libblond().fast_sinv(__getPointer(x), __getLen(x), __getPointer(result))
        return result
    elif isinstance(x, np.ndarray) and isinstance(x[0], np.float32):
        if result is None:
            result = np.empty(len(x), dtype=np.float32, order='C')
        get_libblond().fast_sinvf(__getPointer(x), __getLen(x), __getPointer(result))
        return result
    elif isinstance(x, float) or isinstance(x, np.float32) or isinstance(x, int):
        get_libblond().fast_sin.restype = ct.c_double
        return get_libblond().fast_sin(ct.c_double(x))
    else:
        # TypeError
        raise RuntimeError('[sin] The type %s is not supported' % type(x))


def cos_cpp(x, result=None):
    if isinstance(x, np.ndarray) and isinstance(x[0], np.float64):
        if result is None:
            result = np.empty(len(x), dtype=np.float64, order='C')
        get_libblond().fast_cosv(__getPointer(x), __getLen(x), __getPointer(result))
        return result
    elif isinstance(x, np.ndarray) and isinstance(x[0], np.float32):
        if result is None:
            result = np.empty(len(x), dtype=np.float32, order='C')
        get_libblond().fast_cosvf(__getPointer(x), __getLen(x), __getPointer(result))
        return result
    elif isinstance(x, float) or isinstance(x, np.float32) or isinstance(x, int):
        get_libblond().fast_cos.restype = ct.c_double
        return get_libblond().fast_cos(ct.c_double(x))
    else:
        # TypeError
        raise RuntimeError('[cos] The type %s is not supported' % type(x))


def exp_cpp(x, result=None):
    if isinstance(x, np.ndarray) and isinstance(x[0], np.float64):
        if result is None:
            result = np.empty(len(x), dtype=np.float64, order='C')
        get_libblond().fast_expv(__getPointer(x), __getLen(x), __getPointer(result))
        return result
    elif isinstance(x, np.ndarray) and isinstance(x[0], np.float32):
        if result is None:
            result = np.empty(len(x), dtype=np.float32, order='C')
        get_libblond().fast_expvf(__getPointer(x), __getLen(x), __getPointer(result))
        return result
    elif isinstance(x, float) or isinstance(x, np.float32) or isinstance(x, int):
        get_libblond().fast_exp.restype = ct.c_double
        return get_libblond().fast_exp(ct.c_double(x))
    else:
        # TypeError
        raise RuntimeError('[exp] The type %s is not supported' % type(x))


def interp_cpp(x, xp, yp, left=None, right=None, result=None):
    x = x.astype(dtype=precision.real_t, order='C', copy=False)
    xp = xp.astype(dtype=precision.real_t, order='C', copy=False)
    yp = yp.astype(dtype=precision.real_t, order='C', copy=False)

    if not left:
        left = yp[0]
    if not right:
        right = yp[-1]
    if result is None:
        result = np.empty(len(x), dtype=precision.real_t, order='C')

    get_libblond().interp(__getPointer(x), __getLen(x),
                          __getPointer(xp), __getLen(xp),
                          __getPointer(yp),
                          c_real(left),
                          c_real(right),
                          __getPointer(result))

    return result


def interp_const_space(x, xp, yp, left=None, right=None, result=None):
    x = x.astype(dtype=precision.real_t, order='C', copy=False)
    xp = xp.astype(dtype=precision.real_t, order='C', copy=False)
    yp = yp.astype(dtype=precision.real_t, order='C', copy=False)

    if not left:
        left = yp[0]
    if not right:
        right = yp[-1]
    if result is None:
        result = np.empty(len(x), dtype=precision.real_t, order='C')

    get_libblond().interp_const_space(__getPointer(x), __getLen(x),
                                      __getPointer(xp), __getLen(xp),
                                      __getPointer(yp),
                                      c_real(left),
                                      c_real(right),
                                      __getPointer(result))

    return result


def interp_const_bin(x, xp, yp, left=None, right=None, result=None):
    x = x.astype(dtype=precision.real_t, order='C', copy=False)
    xp = xp.astype(dtype=precision.real_t, order='C', copy=False)
    yp = yp.astype(dtype=precision.real_t, order='C', copy=False)

    if not left:
        left = yp[0]
    if not right:
        right = yp[-1]
    if result is None:
        result = np.empty(len(x), dtype=precision.real_t, order='C')

    get_libblond().interp_const_bin(__getPointer(x), __getLen(x),
                                        __getPointer(xp),
                                        __getPointer(yp), __getLen(xp),
                                        c_real(left), c_real(right),
                                        __getPointer(result))

    return result


def random_normal(loc=0.0, scale=1.0, size=1, seed=1234):
    arr = np.empty(size, dtype=precision.real_t)
    get_libblond().random_normal(
            __getPointer(arr),
            ct.c_double(loc),
            ct.c_double(scale),
            # c_real(loc),
            # c_real(scale),
            __getLen(arr),
            ct.c_ulong(seed))

    return arr


def rfft(a, n=0, result=None):
    a = a.astype(dtype=precision.real_t, order='C', copy=False)
    if (n == 0) and (result is None):
        result = np.empty(
            len(a) // 2 + 1, dtype=precision.complex_t, order='C')
    elif (n != 0) and (result is None):
        result = np.empty(n // 2 + 1, dtype=precision.complex_t, order='C')

    get_libblond().rfft(__getPointer(a),
                        __getLen(a),
                        __getPointer(result),
                        ct.c_int(int(n)),
                        ct.c_int(int(os.environ.get('OMP_NUM_THREADS', 1))))

    return result


def irfft(a, n=0, result=None):
    a = a.astype(dtype=precision.complex_t, order='C', copy=False)

    if (n == 0) and (result is None):
        result = np.empty(2 * (len(a) - 1), dtype=precision.real_t, order='C')
    elif (n != 0) and (result is None):
        result = np.empty(n, dtype=precision.real_t, order='C')

    get_libblond().irfft(__getPointer(a),
                         __getLen(a),
                         __getPointer(result),
                         ct.c_int(int(n)),
                         ct.c_int(int(os.environ.get('OMP_NUM_THREADS', 1))))

    return result


def rfftfreq(n, d=1.0, result=None):
    if d == 0:
        raise ZeroDivisionError('d must be non-zero')
    if result is None:
        result = np.empty(n // 2 + 1, dtype=precision.real_t)

    get_libblond().rfftfreq(ct.c_int(n),
                            __getPointer(result),
                            c_real(d))

    return result


def irfft_packed(signal, fftsize=0, result=None):

    n0 = len(signal[0])
    howmany = len(signal)

    signal = np.ascontiguousarray(np.reshape(
        signal, -1), dtype=precision.complex_t)

    if (fftsize == 0) and (result is None):
        result = np.empty(howmany * 2 * (n0 - 1), dtype=precision.real_t)
    elif (fftsize != 0) and (result is None):
        result = np.empty(howmany * fftsize, dtype=precision.real_t)

    get_libblond().irfft_packed(__getPointer(signal),
                                ct.c_int(n0),
                                ct.c_int(howmany),
                                __getPointer(result),
                                ct.c_int(int(fftsize)),
                                ct.c_int(int(os.environ.get('OMP_NUM_THREADS', 1))))

    result = np.reshape(result, (howmany, -1))

    return result


def cumtrapz(y, x=None, dx=1.0, initial=None, result=None):
    if x is not None:
        # IntegrationError
        raise RuntimeError('[cumtrapz] x attribute is not yet supported')
    if initial:
        if result is None:
            result = np.empty(len(y), dtype=float)
        get_libblond().cumtrapz_w_initial(__getPointer(y),
                                          c_real(dx), c_real(initial),
                                          __getLen(y), __getPointer(result))
    else:
        if result is None:
            result = np.empty(len(y) - 1, dtype=float)
        get_libblond().cumtrapz_wo_initial(__getPointer(y), c_real(dx),
                                           __getLen(y), __getPointer(result))
    return result


def trapz_cpp(y, x=None, dx=1.0):
    if x is None:
        get_libblond().trapz_const_delta.restype = precision.c_real_t
        return get_libblond().trapz_const_delta(__getPointer(y), c_real(dx),
                                                __getLen(y))
    else:
        get_libblond().trapz_var_delta.restype = precision.c_real_t
        return get_libblond().trapz_var_delta(__getPointer(y), __getPointer(x),
                                              __getLen(y))


def beam_phase(bin_centers, profile, alpha, omegarf, phirf, bin_size):
    bin_centers = bin_centers.astype(dtype=precision.real_t, order='C',
                                     copy=False)
    profile = profile.astype(dtype=precision.real_t, order='C', copy=False)

    get_libblond().beam_phase.restype = precision.c_real_t
    coeff = get_libblond().beam_phase(__getPointer(bin_centers),
                                      __getPointer(profile),
                                      c_real(alpha),
                                      c_real(omegarf),
                                      c_real(phirf),
                                      c_real(bin_size),
                                      __getLen(profile))
    return coeff


def beam_phase_fast(bin_centers, profile, omegarf, phirf, bin_size):
    bin_centers = bin_centers.astype(dtype=precision.real_t, order='C',
                                     copy=False)
    profile = profile.astype(dtype=precision.real_t, order='C', copy=False)

    get_libblond().beam_phase_fast.restype = precision.c_real_t
    coeff = get_libblond().beam_phase_fast(__getPointer(bin_centers),
                                           __getPointer(profile),
                                           c_real(omegarf),
                                           c_real(phirf),
                                           c_real(bin_size),
                                           __getLen(profile))
    return coeff


def rf_volt_comp(voltages, omega_rf, phi_rf, bin_centers):

    bin_centers = bin_centers.astype(
        dtype=precision.real_t, order='C', copy=False)
    voltages = voltages.astype(dtype=precision.real_t, order='C', copy=False)
    omega_rf = omega_rf.astype(dtype=precision.real_t, order='C', copy=False)
    phi_rf = phi_rf.astype(dtype=precision.real_t, order='C', copy=False)

    rf_voltage = np.zeros(len(bin_centers), dtype=precision.real_t, order='C')

    get_libblond().rf_volt_comp(__getPointer(voltages),
                                __getPointer(omega_rf),
                                __getPointer(phi_rf),
                                __getPointer(bin_centers),
                                __getLen(voltages),
                                __getLen(rf_voltage),
                                __getPointer(rf_voltage))

    return rf_voltage


def kick(dt, dE, voltage, omega_rf, phi_rf, charge, n_rf, acceleration_kick):
    assert isinstance(dt[0], precision.real_t)
    assert isinstance(dE[0], precision.real_t)

    voltage_kick = charge * \
        voltage.astype(dtype=precision.real_t, order='C', copy=False)
    omegarf_kick = omega_rf.astype(
        dtype=precision.real_t, order='C', copy=False)
    phirf_kick = phi_rf.astype(dtype=precision.real_t, order='C', copy=False)

    get_libblond().kick(__getPointer(dt),
                        __getPointer(dE),
                        ct.c_int(n_rf),
                        __getPointer(voltage_kick),
                        __getPointer(omegarf_kick),
                        __getPointer(phirf_kick),
                        __getLen(dt),
                        c_real(acceleration_kick))


def drift(dt, dE, solver, t_rev, length_ratio, alpha_order, eta_0,
          eta_1, eta_2, alpha_0, alpha_1, alpha_2, beta, energy):
    assert isinstance(dt[0], precision.real_t)
    assert isinstance(dE[0], precision.real_t)

    solver_to_int = {
        'simple': 0,
        'legacy': 1,
        'exact': 2,
    }
    solver = solver_to_int[solver]

    get_libblond().drift(__getPointer(dt), __getPointer(dE),
                         ct.c_int(solver), c_real(t_rev),
                         c_real(length_ratio), c_real(alpha_order),
                         c_real(eta_0), c_real(eta_1),
                         c_real(eta_2), c_real(alpha_0),
                         c_real(alpha_1), c_real(alpha_2),
                         c_real(beta), c_real(energy),
                         __getLen(dt))


def linear_interp_kick(dt, dE, voltage,
                       bin_centers, charge,
                       acceleration_kick):

    assert isinstance(dt[0], precision.real_t)
    assert isinstance(dE[0], precision.real_t)
    assert isinstance(voltage[0], precision.real_t)
    assert isinstance(bin_centers[0], precision.real_t)

    get_libblond().linear_interp_kick(__getPointer(dt),
                                      __getPointer(dE),
                                      __getPointer(voltage),
                                      __getPointer(bin_centers),
                                      c_real(charge),
                                      __getLen(bin_centers),
                                      __getLen(dt),
                                      c_real(acceleration_kick))


def slice_beam(dt, profile, cut_left, cut_right):
    assert isinstance(dt[0], precision.real_t)
    assert isinstance(profile[0], precision.real_t)

    get_libblond().histogram(__getPointer(dt),
                             __getPointer(profile),
                             c_real(cut_left),
                             c_real(cut_right),
                             __getLen(profile),
                             __getLen(dt))


def slice_smooth(dt, profile, cut_left, cut_right):
    assert isinstance(dt[0], precision.real_t)
    assert isinstance(profile[0], precision.real_t)

    get_libblond().smooth_histogram(__getPointer(dt),
                                    __getPointer(profile),
                                    c_real(cut_left),
                                    c_real(cut_right),
                                    __getLen(profile),
                                    __getLen(dt))


def sparse_histogram(dt, profile, cut_left, cut_right, bunch_indexes, n_slices_bucket):
    assert isinstance(dt[0], precision.real_t)
    assert isinstance(profile[0][0], precision.real_t)

    get_libblond().sparse_histogram(__getPointer(dt),
                                    __getPointer(profile),
                                    __getPointer(cut_left),
                                    __getPointer(cut_right),
                                    __getPointer(bunch_indexes),
                                    ct.c_int(n_slices_bucket),
                                    __getLen(cut_left),
                                    __getLen(dt))


def music_track(dt, dE, induced_voltage, array_parameters,
                alpha, omega_bar,
                const, coeff1, coeff2, coeff3, coeff4):
    assert isinstance(dt[0], precision.real_t)
    assert isinstance(dE[0], precision.real_t)
    assert isinstance(induced_voltage[0], precision.real_t)
    assert isinstance(array_parameters[0], precision.real_t)

    get_libblond().music_track(__getPointer(dt),
                               __getPointer(dE),
                               __getPointer(induced_voltage),
                               __getPointer(array_parameters),
                               __getLen(dt),
                               c_real(alpha),
                               c_real(omega_bar),
                               c_real(const),
                               c_real(coeff1),
                               c_real(coeff2),
                               c_real(coeff3),
                               c_real(coeff4))


def music_track_multiturn(dt, dE, induced_voltage, array_parameters,
                          alpha, omega_bar,
                          const, coeff1, coeff2, coeff3, coeff4):
    assert isinstance(dt[0], precision.real_t)
    assert isinstance(dE[0], precision.real_t)
    assert isinstance(induced_voltage[0], precision.real_t)
    assert isinstance(array_parameters[0], precision.real_t)

    get_libblond().music_track_multiturn(__getPointer(dt),
                                         __getPointer(dE),
                                         __getPointer(induced_voltage),
                                         __getPointer(array_parameters),
                                         __getLen(dt),
                                         c_real(alpha),
                                         c_real(omega_bar),
                                         c_real(const),
                                         c_real(coeff1),
                                         c_real(coeff2),
                                         c_real(coeff3),
                                         c_real(coeff4))


def synchrotron_radiation(dE, U0, n_kicks, tau_z):
    assert isinstance(dE[0], precision.real_t)

    get_libblond().synchrotron_radiation(
        __getPointer(dE),
        c_real(U0),
        __getLen(dE),
        c_real(tau_z),
        ct.c_int(n_kicks))


def synchrotron_radiation_full(dE, U0, n_kicks, tau_z, sigma_dE, energy):
    assert isinstance(dE[0], precision.real_t)

    get_libblond().synchrotron_radiation_full(
        __getPointer(dE),
        c_real(U0),
        __getLen(dE),
        c_real(sigma_dE),
        c_real(tau_z),
        c_real(energy),
        ct.c_int(n_kicks))


def set_random_seed(seed):
    get_libblond().set_random_seed(ct.c_int(seed))


def fast_resonator(R_S, Q, frequency_array, frequency_R, impedance=None):
    R_S = R_S.astype(dtype=precision.real_t, order='C', copy=False)
    Q = Q.astype(dtype=precision.real_t, order='C', copy=False)
    frequency_array = frequency_array.astype(
        dtype=precision.real_t, order='C', copy=False)
    frequency_R = frequency_R.astype(
        dtype=precision.real_t, order='C', copy=False)

    realImp = np.zeros(len(frequency_array), dtype=precision.real_t)
    imagImp = np.zeros(len(frequency_array), dtype=precision.real_t)

    get_libblond().fast_resonator_real_imag(
        __getPointer(realImp),
        __getPointer(imagImp),
        __getPointer(frequency_array),
        __getPointer(R_S),
        __getPointer(Q),
        __getPointer(frequency_R),
        __getLen(R_S),
        __getLen(frequency_array))

    impedance = realImp + 1j * imagImp
    return impedance


def distribution_from_tomoscope(dt, dE, probDistr, seed, profLen,
                                cutoff, x0, y0, dtBin, dEBin):

    get_libblond().generate_distribution(__getPointer(dt),
                                         __getPointer(dE),
                                         __getPointer(probDistr),
                                         ct.c_uint(seed), ct.c_uint(profLen),
                                         c_real(cutoff), c_real(x0),
                                         c_real(y0), c_real(dtBin),
                                         c_real(dEBin), __getLen(dt))



# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public License version 3 (GPL Version 3),
# copied verbatim in the file LICENSE.md.
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Function(s) for pre-processing input data**

:Authors: **Simon Albright**
'''

import numpy as np

import blond.utils.exceptions as blExcept


def check_input(variable, msg, *args):
    """Check input and return InputDataError exception with user defined
    message if False

    Args:
        variable (_type_): _description_
        msg (_type_): _description_

    Raises:
        blExcept.InputDataError: _description_

    Returns:
        _type_: _description_
    """
    result = check_data_dimensions(variable, *args)

    if result[0]:
        return result
    raise blExcept.InputDataError(msg)


def check_data_dimensions(input_data, *args):
    """
    General function to check if input_data is number, or nD array
    for each member of args the input_data is checked
    0 is taken as checking if input_data is a number
    integers are taken as the length of a list-like input
    tuples or lists are taken as dimensions
    [[1, 2], [1, 2]] will return true for length = 2 or dims = (2, 2)

    Args:
        input_data (_type_): _description_

    Returns:
        _type_: _description_
    """
    success = False

    for arg in args:
        if arg == 0:
            if _check_number(input_data):
                success = True

        elif isinstance(arg, int):
            if _check_length(input_data, arg):
                success = True

        else:
            if _check_dimensions(input_data, arg):
                success = True

        if success:
            break

    return success, type(input_data)


def _check_number(input_data):
    """returns True if input_data can be cast to int

    Args:
        input_data (_type_): _description_

    Raises:
        TypeError: _description_

    Returns:
        _type_: _description_
    """
    try:
        int(input_data)
        if isinstance(input_data, np.ndarray):
            raise TypeError
        return True
    except (TypeError, ValueError):
        return False


def _check_length(input_data, length):
    """ Returns True if len(input_data) == length
    Should this return True if n-dim > 1?

    Args:
        input_data (_type_): _description_
        length (_type_): _description_

    Raises:
        TypeError: _description_

    Returns:
        _type_: _description_
    """
    if not _check_number(length):
        raise TypeError("Length must be numeric")

    try:
        return len(input_data) == length
    except TypeError:
        return False


def _check_dimensions(input_data, dim):
    """
    Casts input_data to numpy array and dimensions to tuple
    compares shape of array to tuple and returns True if equal.
    If dim[n] == -1 it is set to input_data.shape[n] to allow comprehension
    of arrays with arbitrary length in one or more dimensions.
    Args:
        input_data (_type_): _description_
        dim (_type_): _description_

    Raises:
        TypeError: _description_

    Returns:
        _type_: _description_
    """
    try:
        iter(dim)
    except TypeError:
        dim = [dim]

    input_shape = np.array(input_data).shape

    try:
        if -1 in dim:
            try:
                dim = [input_shape[i] if dim[i] == -1
                       else dim[i] for i in range(len(dim))]
            except IndexError:
                return False
    except TypeError as exc:
        raise TypeError("dim must be number or iterable of numbers") from exc

    return input_shape == tuple(dim)




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Functions to plot general and RF paramaters**

:Authors: **Helga Timko**
'''

from __future__ import division

import matplotlib.pyplot as plt


def plot_voltage_programme(time, voltage, sampling=1, dirname='fig',
                           figno=0):
    """
    Plot of the RF voltage as a function of time.
    For large amount of data, use "sampling" to plot a fraction of the data.
    """

    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8, 6)
    ax = plt.axes([0.15, 0.1, 0.8, 0.8])
    ax.plot(time[::sampling], voltage[::sampling])
    ax.set_xlabel("Time [s]")
    ax.set_ylabel(r"RF voltage [V]")

    # Save figure
    fign = dirname + '/RF_voltage_' "%d" % figno + '.png'
    plt.savefig(fign)
    plt.clf()



# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Various beam phase loops with optional synchronisation/frequency/radial loops
for the CERN machines**

:Authors: **Helga Timko**, **Alexandre Lasheen**
'''

from __future__ import division


import numpy as np

from ..utils import bmath as bm


class BeamFeedback:
    '''
    One-turn beam phase loop for different machines with different hardware. 
    Use 'period' for a phase loop that is active only in certain turns. 
    The phase loop acts directly on the RF frequency of all harmonics and
    affects the RF phase as well.
    '''

    def __init__(self, Ring, RFStation, Profile,
                 configuration,
                 PhaseNoise=None,
                 LHCNoiseFB=None, delay=0):

        #: | *Import Ring*
        self.ring = Ring

        #: | *Import RFStation*
        self.rf_station = RFStation

        #: | *Import Profile*
        self.profile = Profile

        #: | *Machine-dependent configuration of LLRF system.*
        self.config = configuration

        self.delay = delay

        #: | *Machine name; see description of each machine.*
        if 'machine' not in self.config:
            self.machine = 'LHC'
        else:
            self.machine = self.config['machine']

        #: | *Band-pass filter window coefficient for beam phase calculation.*
        if 'window_coefficient' not in self.config:
            self.alpha = 0.
        else:
            self.alpha = self.config['window_coefficient']

        # determines from which RF-buckets the band-pass filter starts to acts
        if 'time_offset' not in self.config:
            self.time_offset = None
        else:
            self.time_offset = self.config['time_offset']

        #: | *Phase loop gain. Implementation depends on machine.*
        try:
            self.gain = self.config['PL_gain']
        except Exception:
            # PhaseLoopError
            raise RuntimeError(
                "You need to specify the Phase Loop gain! Aborting")

        # LHC CONFIGURATION
        if self.machine == 'LHC':

            #: | *Synchronisation loop gain.*
            if 'SL_gain' not in self.config:
                self.gain2 = 0.
            else:
                self.gain2 = self.config['SL_gain']

            #: | *LHC Synchroronisation loop recursion variable*
            self.lhc_y = 0

            if self.gain2 != 0:

                #: | *LHC Synchronisation loop coefficient [1]*
                self.lhc_a = 5.25 - self.rf_station.omega_s0 / (np.pi * 40.)
                #: | *LHC Synchronisation loop time constant [turns]*
                self.lhc_t = (2 * np.pi * self.rf_station.Q_s * np.sqrt(self.lhc_a)) / \
                    np.sqrt(1 + self.gain / self.gain2 *
                            np.sqrt((1 + 1 / self.lhc_a) / (1 + self.lhc_a)))

            else:

                self.lhc_a = np.zeros(self.rf_station.n_turns + 1)
                self.lhc_t = np.zeros(self.rf_station.n_turns + 1)

        # LHC_F CONFIGURATION
        elif self.machine == 'LHC_F':

            #: | *Frequency loop gain.*
            if 'FL_gain' not in self.config:
                self.gain2 = 0.
            else:
                self.gain2 = self.config['FL_gain']

        # SPS_RL CONFIGURATION
        elif self.machine == 'SPS_RL':

            #: | *Frequency loop gain.*
            if 'RL_gain' not in self.config:
                self.gain2 = 0.
            else:
                self.gain2 = self.config['RL_gain']

            #: | *Number of particles to sample from dE for orbit calculation*
            if 'sample_dE' not in self.config:
                self.sample_dE = 1
            else:
                self.sample_dE = self.config['sample_dE']

        elif self.machine == 'SPS_F':
            #: | *Frequency loop gain.*
            if 'FL_gain' not in self.config:
                self.gain2 = 0.
            else:
                self.gain2 = self.config['FL_gain']

        # PSB CONFIGURATION
        elif self.machine == 'PSB':

            self.gain = self.gain * np.ones(Ring.n_turns + 1)

            #: | *Radial loop gain, proportional [1] and integral [1/s].*
            if 'RL_gain' not in self.config:
                self.gain2 = [0., 0.]
            else:
                self.gain2 = self.config['RL_gain']

            self.gain2[0] = self.gain2[0] * np.ones(Ring.n_turns + 1)
            self.gain2[1] = self.gain2[1] * np.ones(Ring.n_turns + 1)

            #: | *Optional: PL & RL acting only in certain time intervals/turns.*
            self.dt = 0
            # | *Phase Loop sampling period [s]*
            if 'period' not in self.config:
                self.dt = 10.e-6  # [s]
            else:
                self.dt = self.config['period']

            # Counter of turns passed since last time the PL was active
            self.PL_counter = 0
            self.on_time = np.array([])

            self.precalculate_time(Ring)

            #: | *Array of transfer function coefficients.*
            if 'coefficients' not in self.config:
                self.coefficients = [0.999019, -0.999019, 0., 1., -0.998038, 0.]
            else:
                self.coefficients = self.config['coefficients']

            #: | *Memory of previous phase correction, for phase loop.*
            self.dphi_sum = 0.
            self.dphi_av = 0.
            self.dphi_av_prev = 0.

            #: | *Memory of previous relative radial correction, for rad loop.*
            self.dR_over_R_prev = 0.

            #: | *Phase loop frequency correction [1/s]*
            self.domega_PL = 0.

            #: | *Radial loop frequency correction [1/s]*
            self.domega_RL = 0.

            self.dR_over_R = 0

        #: | *Relative radial displacement [1], for radial loop.*
        self.drho = 0.

        #: | *Phase loop frequency correction of the main RF system.*
        self.domega_rf = 0.

        #: | *Beam phase measured at the main RF frequency.*
        self.phi_beam = 0.

        #: | *Phase difference between beam and RF.*
        self.dphi = 0.

        #: | *Reference signal for secondary loop to test step response.*
        self.reference = 0.

        #: | *Optional import of RF PhaseNoise object*
        self.RFnoise = PhaseNoise
        if (self.RFnoise is not None
                and (len(self.RFnoise.dphi) != Ring.n_turns + 1)):
            # PhaseNoiseError
            raise RuntimeError(
                'Phase noise has to have a length of n_turns + 1')

        #: | *Optional import of amplitude-scaling feedback object LHCNoiseFB*
        self.noiseFB = LHCNoiseFB

    def track(self):
        '''
        Calculate PL correction on main RF frequency depending on machine and
        propagate it to other RF systems.
        The update of the RF phase and frequency for the next turn,
        for all systems is done in the tracker.
        '''

        # Calculate PL correction on RF frequency
        getattr(self, self.machine)()

        # Update the RF frequency of all systems for the next turn
        counter = self.rf_station.counter[0] + 1
        self.rf_station.omega_rf[:, counter] += self.domega_rf * \
            self.rf_station.harmonic[:, counter] / \
            self.rf_station.harmonic[0, counter]

    def precalculate_time(self, Ring):
        '''
        *For machines like the PSB, where the PL acts only in certain time
        intervals, pre-calculate on which turns to act.*
        '''

        if self.dt > 0:
            n = self.delay + 1
            while n < Ring.t_rev.size:
                summa = 0
                while summa < self.dt:
                    try:
                        summa += Ring.t_rev[n]
                        n += 1
                    except Exception:
                        self.on_time = np.append(self.on_time, 0)
                        return
                self.on_time = np.append(self.on_time, n - 1)
        else:
            self.on_time = np.arange(Ring.t_rev.size)

    def beam_phase(self):
        '''
        *Beam phase measured at the main RF frequency and phase. The beam is 
        convolved with the window function of the band-pass filter of the 
        machine. The coefficients of sine and cosine components determine the 
        beam phase, projected to the range -Pi/2 to 3/2 Pi. Note that this beam
        phase is already w.r.t. the instantaneous RF phase.*
        '''

        # Main RF frequency at the present turn
        omega_rf = self.rf_station.omega_rf[0, self.rf_station.counter[0]]
        phi_rf = self.rf_station.phi_rf[0, self.rf_station.counter[0]]

        if self.time_offset is None:
            coeff = bm.beam_phase(self.profile.bin_centers,
                                  self.profile.n_macroparticles,
                                  self.alpha, omega_rf, phi_rf,
                                  self.profile.bin_size)
        else:
            indexes = self.profile.bin_centers >= self.time_offset
            coeff = bm.beam_phase(self.profile.bin_centers[indexes],
                                  self.profile.n_macroparticles[indexes],
                                  self.alpha, omega_rf, phi_rf,
                                  self.profile.bin_size)

        # Project beam phase to (pi/2,3pi/2) range
        self.phi_beam = np.arctan(coeff) + np.pi

    def beam_phase_sharpWindow(self):
        '''
        *Beam phase measured at the main RF frequency and phase. The beam is
        averaged over a window. The coefficients of sine and cosine components
        determine the beam phase, projected to the range -Pi/2 to 3/2 Pi.
        Note that this beam phase is already w.r.t. the instantaneous RF phase.*
        '''

        # Main RF frequency at the present turn
        turn = self.rf_station.counter[0]
        omega_rf = self.rf_station.omega_rf[0, turn]
        phi_rf = self.rf_station.phi_rf[0, turn]

        if self.alpha != 0.0:
            indexes = bm.logical_and((self.time_offset - np.pi / omega_rf)
                                     <= self.profile.bin_centers,
                                     self.profile.bin_centers
                                     <= (-1 / self.alpha + self.time_offset -
                                         2 * np.pi / omega_rf))
        else:
            indexes = bm.ones(self.profile.n_slices, dtype=bool)

        # Convolve with window function
        scoeff = bm.trapz(bm.sin(omega_rf * self.profile.bin_centers[indexes]
                                 + phi_rf)
                          * self.profile.n_macroparticles[indexes],
                          dx=self.profile.bin_size)
        ccoeff = bm.trapz(bm.cos(omega_rf * self.profile.bin_centers[indexes]
                                 + phi_rf) *
                          self.profile.n_macroparticles[indexes],
                          dx=self.profile.bin_size)

        # Project beam phase to (pi/2,3pi/2) range
        self.phi_beam = np.arctan(scoeff / ccoeff) + np.pi

    def phase_difference(self):
        '''
        *Phase difference between beam and RF phase of the main RF system.
        Optional: add RF phase noise through dphi directly.*
        '''

        # Correct for design stable phase
        counter = self.rf_station.counter[0]
        self.dphi = self.phi_beam - self.rf_station.phi_s[counter]

        # Possibility to add RF phase noise through the PL
        if self.RFnoise is not None:
            if self.noiseFB is not None:
                self.dphi += self.noiseFB.x * self.RFnoise.dphi[counter]
            else:
                if self.machine == 'PSB':
                    self.dphi = self.dphi
                else:
                    self.dphi += self.RFnoise.dphi[counter]

    def radial_difference(self):
        '''
        *Radial difference between beam and design orbit.*
        '''

        counter = self.rf_station.counter[0]

        # Correct for design orbit
#        self.average_dE = np.mean(self.profile.Beam.dE[(self.profile.Beam.dt >
#            self.profile.bin_centers[0])*(self.profile.Beam.dt <
#                                         self.profile.bin_centers[-1])])
        self.average_dE = bm.mean(self.profile.Beam.dE[::self.sample_dE])

        self.drho = self.ring.alpha_0[0, counter] * \
            self.ring.ring_radius * self.average_dE / \
            (self.ring.beta[0, counter]**2.
             * self.ring.energy[0, counter])

    def radial_steering_from_freq(self):
        '''
        *Frequency and phase change for the current turn due to the radial steering program.*
        '''

        counter = self.rf_station.counter[0]

        self.radial_steering_domega_rf = - self.rf_station.omega_rf_d[0, counter] * \
            self.rf_station.eta_0[counter] / self.ring.alpha_0[0, counter] * \
            self.reference / self.ring.ring_radius

        self.rf_station.omega_rf[:, counter] += self.radial_steering_domega_rf * \
            self.rf_station.harmonic[:, counter] / \
            self.rf_station.harmonic[0, counter]

        # Update the RF phase of all systems for the next turn
        # Accumulated phase offset due to PL in each RF system
        self.rf_station.dphi_rf_steering += 2. * np.pi * self.rf_station.harmonic[:, counter] * \
            (self.rf_station.omega_rf[:, counter] -
             self.rf_station.omega_rf_d[:, counter]) / \
            self.rf_station.omega_rf_d[:, counter]

        # Total phase offset
        self.rf_station.phi_rf[:, counter] += self.rf_station.dphi_rf_steering

    def LHC_F(self):
        '''
        Calculation of the LHC RF frequency correction from the phase difference
        between beam and RF (actual synchronous phase). The transfer function is

        .. math::
            \\Delta \\omega_{rf}^{PL} = - g_{PL} (\\Delta\\varphi_{PL} + \\phi_{N}) 

        where the phase noise for the controlled blow-up can be optionally 
        activated.  
        Using 'gain2', a frequency loop can be activated in addition to remove
        long-term frequency drifts:

        .. math::
            \\Delta \\omega_{rf}^{FL} = - g_{FL} (\\omega_{rf} - h \\omega_{0})    
        '''

        counter = self.rf_station.counter[0]

        self.beam_phase()
        self.phase_difference()

        # Frequency correction from phase loop and frequency loop
        self.domega_rf = - self.gain * self.dphi \
            - self.gain2 * (self.rf_station.omega_rf[0, counter] -
                            self.rf_station.omega_rf_d[0, counter] +
                            self.reference)

    def SPS_F(self):
        '''
        Calculation of the SPS RF frequency correction from the phase
        difference between beam and RF (actual synchronous phase). Same as 
        LHC_F, except the calculation of the beam phase.
        '''

        counter = self.rf_station.counter[0]

        self.beam_phase_sharpWindow()
        self.phase_difference()

        # Frequency correction from phase loop and frequency loop
        self.domega_dphi = - self.gain * self.dphi
        self.domega_df = - self.gain2 * (self.rf_station.omega_rf[0, counter] -
                                         self.rf_station.omega_rf_d[0, counter])

        self.domega_rf = self.domega_dphi + self.domega_df

    def SPS_RL(self):
        '''
        Calculation of the SPS RF frequency correction from the phase difference
        between beam and RF (actual synchronous phase). The transfer function is

        .. math::
            \\Delta \\omega_{rf}^{PL} = - g_{PL} (\\Delta\\varphi_{PL} + \\phi_{N}) 

        where the phase noise for the controlled blow-up can be optionally 
        activated.  
        Using 'gain2', a radial loop can be activated in addition to remove
        long-term frequency drifts
        '''

        counter = self.rf_station.counter[0]

        if self.reference != 0:
            self.radial_steering_from_freq()

        self.beam_phase()
        self.phase_difference()
        self.radial_difference()

        # Frequency correction from phase loop and radial loop
        self.domega_dphi = - self.gain * self.dphi
        self.domega_dR = - bm.sign(self.rf_station.eta_0[counter]) * self.gain2 * \
            (self.reference - self.drho) / self.ring.ring_radius

        self.domega_rf = self.domega_dphi + self.domega_dR

    def LHC(self):
        '''
        Calculation of the LHC RF frequency correction from the phase difference
        between beam and RF (actual synchronous phase). The transfer function is

        .. math::
            \\Delta \\omega_{rf}^{PL} = - g_{PL} (\\Delta\\varphi_{PL} + \\phi_{N}) 

        where the phase noise for the controlled blow-up can be optionally 
        activated.  
        Using 'gain2', a synchro loop can be activated in addition to remove
        long-term frequency drifts:     

        .. math::
            \\Delta \\omega_{rf}^{SL} = - g_{SL} (y + a \\Delta\\varphi_{rf}) ,

        where we use the recursion

        .. math::
            y_{n+1} = (1 - \\tau) y_n + (1 - a) \\tau \\Delta\\varphi_{rf} ,

        with a and \tau being defined through the synchrotron frequency f_s and
        the synchrotron tune Q_s as

        .. math::
            a (f_s) \\equiv 5.25 - \\frac{f_s}{\\pi 40~\\text{Hz}} ,

        .. math::
            \\tau(f_s) \\equiv 2 \\pi Q_s \\sqrt{ \\frac{a}{1 + \\frac{g_{PL}}{g_{SL}} \\sqrt{\\frac{1 + 1/a}{1 + a}} }}
        '''

        counter = self.rf_station.counter[0]
        dphi_rf = self.rf_station.dphi_rf[0]

        self.beam_phase()
        self.phase_difference()

        # Frequency correction from phase loop and synchro loop
        self.domega_rf = - self.gain * self.dphi \
                         - self.gain2 * (self.lhc_y + self.lhc_a[counter]
                                         * (dphi_rf + self.reference))

        # Update recursion variable
        self.lhc_y = (1 - self.lhc_t[counter]) * self.lhc_y + \
                     (1 - self.lhc_a[counter]) * self.lhc_t[counter] * \
                     (dphi_rf + self.reference)

    def PSB(self):
        '''
        Phase and radial loops for PSB. See documentation on-line for details.
        '''

        # Average phase error while frequency is updated
        counter = self.rf_station.counter[0]
        self.beam_phase()
        self.phase_difference()

        self.dphi_sum += self.dphi

        # Phase and radial loop active on certain turns
        if counter == self.on_time[self.PL_counter] and counter >= self.delay:
            # Phase loop
            self.dphi_av = self.dphi_sum / (self.on_time[self.PL_counter]
                                            - self.on_time[self.PL_counter - 1])

            if self.RFnoise is not None:
                self.dphi_av += self.RFnoise.dphi[counter]

            self.domega_PL = 0.99803799 * self.domega_PL \
                + self.gain[counter] * (0.99901903 * self.dphi_av -
                                        0.99901003 * self.dphi_av_prev)

            self.dphi_av_prev = self.dphi_av
            self.dphi_sum = 0.

            # Radial loop
            self.dR_over_R = (self.rf_station.omega_rf[0, counter] -
                              self.rf_station.omega_rf_d[0, counter]) / (
                self.rf_station.omega_rf_d[0, counter] *
                (1. / (self.ring.alpha_0[0, counter] *
                       self.rf_station.gamma[counter]**2) - 1.))

            self.domega_RL = self.domega_RL + self.gain2[0][counter] * (self.dR_over_R
                                                                        - self.dR_over_R_prev) + self.gain2[1][counter] * self.dR_over_R

            self.dR_over_R_prev = self.dR_over_R

            # Counter to pick the next time step when the PL & RL will be active
            self.PL_counter += 1

        # Apply frequency correction
        self.domega_rf = - self.domega_PL - self.domega_RL

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return

        # No arrays need to be transfered

        # to make sure it will not be called again
        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return

        # No arrays need to be transfered

        # to make sure it will not be called again
        self._device = 'CPU'



# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Frequency corrections to design frequency to allow fixed injection frequency
and frequency offsets**

:Authors: **Simon Albright**
'''

import numpy as np


class _FrequencyOffset:
    '''
    Compute effect of having a different RF and design frequency
    '''

    def __init__(self, Ring, RFStation, System=None, MainH=None):

        #: | *Import Ring*
        self.ring = Ring

        #: | *Import RFStation*
        self.rf_station = RFStation

        #: | *Set system number(s) to modify, if None all are modified*
        if isinstance(System, int):
            self.system = [System]
        elif hasattr(System, '__iter__'):
            self.system = []
            for s in System:
                self.system.append(s)
        elif System is None:
            self.system = System
        else:
            raise TypeError("System must be int, iterable of ints or None")

        if self.system and not all((isinstance(s, int) for s in self.system)):
            raise TypeError("System must be int, iterable of ints or None")

        #: | *Main harmonic the delta F is taken as being in reference to,
        #: |  if None RFStation.harmonic[0][0] is taken as the main*
        if MainH is not None:
            self.mainH = MainH
        else:
            self.mainH = RFStation.harmonic[0][0]

    def set_frequency(self, NewFrequencyProgram):
        '''
        Set new frequency program
        '''

        #: | *Check of frequency is passed as array of [time, freq]*
        if isinstance(NewFrequencyProgram, np.ndarray):
            if NewFrequencyProgram.shape[0] == 2:
                end_turn = np.where(self.ring.cycle_time >=
                                    NewFrequencyProgram[0][-1])[0][0]
                NewFrequencyProgram = np.interp(self.ring.cycle_time[:end_turn],
                                                NewFrequencyProgram[0], NewFrequencyProgram[1])

        #: | *Store new frequency as numpy array relative to the main harmonic*
        self.new_frequency = np.array(NewFrequencyProgram) / self.mainH

        self.end_turn = len(self.new_frequency)

        #: | *Store design frequency during offset*
        self.design_frequency = self.rf_station.omega_rf_d[:, :self.end_turn]

    def calculate_phase_slip(self):
        '''
        Calculate the phase slippage resulting from the frequency offset for \
        each RF system
        '''

        delta_phi = (2 * np.pi * self.rf_station.harmonic[:, :self.end_turn]
                     * (self.rf_station.harmonic[:, :self.end_turn]
                         * self.new_frequency
                         - self.design_frequency)
                     / self.design_frequency)
        self.phase_slippage = np.cumsum(delta_phi, axis=1)

    def apply_new_frequency(self):
        '''
        Sets the RF frequency and phase
        '''

        if self.system is None:
            self.rf_station.omega_rf[:, :self.end_turn] = \
                (self.rf_station.harmonic[:, :self.end_turn]
                 * self.new_frequency)

            self.rf_station.phi_rf[:, :self.end_turn] += self.phase_slippage

            for n in range(self.rf_station.n_rf):
                self.rf_station.phi_rf[n, self.end_turn:] \
                    += self.phase_slippage[n, -1]

        else:
            for system in self.system:
                self.rf_station.omega_rf[system, :self.end_turn] \
                    = (self.rf_station.harmonic[system, :self.end_turn]
                        * self.new_frequency)
                self.rf_station.phi_rf[system, :self.end_turn] \
                    += self.phase_slippage[system]
                self.rf_station.phi_rf[system, self.end_turn:] \
                    += self.phase_slippage[system, -1]


class FixedFrequency(_FrequencyOffset):
    '''
    Compute effect of fixed RF frequency different to frequency from momentum
    program at the start of the cycle.
    '''

    def __init__(self, Ring, RFStation, FixedFrequency, FixedDuration,
                 TransitionDuration, transition=1):

        _FrequencyOffset.__init__(self, Ring, RFStation)

        #: | *Set value of fixed frequency*
        self.fixed_frequency = FixedFrequency

        #: | *Duration of fixed frequency*
        self.fixed_duration = FixedDuration

        #: | *Duration of transition to design frequency*
        self.transition_duration = TransitionDuration

        self.end_fixed_turn = np.where(self.ring.cycle_time >=
                                       self.fixed_duration)[0][0]
        self.end_transition_turn = np.where(self.ring.cycle_time >=
                                            (self.fixed_duration + self.transition_duration))[0][0]

        self.end_frequency = self.rf_station.omega_rf_d[0, self.end_transition_turn]

        if transition == 1:
            self.calculate_frequency_prog = self.transition_1

        self.compute()

    def compute(self):

        self.calculate_frequency_prog()
        self.set_frequency(self.frequency_prog)
        self.calculate_phase_slip()
        self.apply_new_frequency()

    def linear_calculate_frequency_prog(self):
        '''
        Calculate the fixed and transition frequency programs turn by turn
        '''

        fixed_frequency_prog = np.ones(self.end_fixed_turn) * self.fixed_frequency
        transition_frequency_prog = np.linspace(float(self.fixed_frequency),
                                                float(self.end_frequency),
                                                (self.end_transition_turn
                                                - self.end_fixed_turn))

        self.frequency_prog = np.concatenate((fixed_frequency_prog,
                                             transition_frequency_prog))

    def transition_1(self):

        t1 = (self.ring.cycle_time[self.end_transition_turn]
              - self.ring.cycle_time[self.end_fixed_turn])
        f1 = self.end_frequency
        f1Prime = (np.gradient(self.rf_station.omega_rf_d[0])
                   / np.gradient(self.ring.cycle_time))[self.end_transition_turn]

        constA = (t1 * f1Prime - 2 * (f1 - self.fixed_frequency)) / t1**3
        constB = - (t1 * f1Prime - 3 * (f1 - self.fixed_frequency)) / t1**2

        transTime = (self.ring.cycle_time[self.end_fixed_turn:self.end_transition_turn]
                     - self.ring.cycle_time[self.end_fixed_turn])

        transition_freq = (constA * transTime**3 + constB * transTime**2
                           + self.fixed_frequency)

        self.frequency_prog = np.concatenate((np.ones(self.end_fixed_turn)
                                              * self.fixed_frequency, transition_freq))
